"""
Functions for transforming rasters represented as arrays.
"""
from .convolution_kernels import *
from .distance import *
from .convolution_funcs import *
from .convolution import *
from .edge_detection import *
from .morphology import *
from .patches import *
from .filters import *
from .color import *
from .utils_array import *
from .fill import *
from .timeseries import *
from .loaders import *


"""
### Clip rasters ###

Clips a raster using a vector geometry or the extents of a raster.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, List, Optional
from warnings import warn

# External
from osgeo import gdal, ogr
import numpy as np

# Internal
from buteo.utils import (
    utils_io,
    utils_gdal,
    utils_base,
    utils_bbox,
    utils_path,
    utils_translate,
)
from buteo.raster import core_raster
from buteo.vector import core_vector
from buteo.vector.reproject import _vector_reproject


def _raster_clip(
    raster: Union[str, List[str], gdal.Dataset, List[gdal.Dataset]],
    clip_geom: Union[str, gdal.Dataset, ogr.DataSource, ogr.Layer, List[ogr.Layer], List[ogr.DataSource], List[gdal.Dataset]],
    out_path: Optional[Union[str, List[str]]] = None,
    *,
    resample_alg: str = "nearest",
    crop_to_geom: bool = True,
    adjust_bbox: bool = True,
    all_touch: bool = True,
    to_extent: bool = False,
    overwrite: bool = True,
    creation_options: Optional[List] = None,
    dst_nodata: str = "infer",
    src_nodata: str = "infer",
    layer_to_clip: int = 0,
    prefix: str = "",
    suffix: str = "",
    verbose: int = 1,
    ram: float = 0.8,
    ram_max: Optional[int] = None,
    ram_min: Optional[int] = 100,
):
    """
    INTERNAL.
    Clips a raster(s) using a vector geometry or the extents of a raster.
    """
    path_list = utils_io._get_output_paths(
        raster,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=out_path is None,
    )
    assert utils_path._check_is_valid_output_path_list(path_list, overwrite), (
        f"Unable to parse out_path: {out_path}"
    )

    clip_ds = None

    memory_files = []

    if isinstance(clip_geom, ogr.Geometry):
        clip_geom = utils_bbox._get_vector_from_geom(clip_geom)
        clip_geom = utils_gdal._get_path_from_dataset(clip_geom)

    # Input is a vector.
    if utils_gdal._check_is_vector(clip_geom):
        clip_ds = core_vector._vector_open(clip_geom)

        if clip_ds.GetLayerCount() > 1:
            clip_ds = core_vector.vector_filter_layer(clip_ds, layer_name_or_idx=layer_to_clip, add_uuid=True)
            memory_files.append(clip_ds)

        clip_metadata = core_vector._get_basic_metadata_vector(clip_ds)

        if to_extent:
            clip_ds = utils_bbox._get_vector_from_bbox(clip_metadata["bbox"], clip_metadata["projection_osr"])
            memory_files.append(clip_ds)

        if isinstance(clip_ds, ogr.DataSource):
            clip_ds = clip_ds.GetName()

    # Input is a raster (use extent)
    elif utils_gdal._check_is_raster(clip_geom):
        clip_ds = core_raster.raster_to_extent(clip_geom)
        clip_metadata = core_raster._get_basic_metadata_raster(clip_geom)
        memory_files.append(clip_ds)
    else:
        if utils_path._check_file_exists(clip_geom):
            raise ValueError(f"Unable to parse clip geometry: {clip_geom}")

        raise ValueError(f"Unable to locate clip geometry {clip_geom}")

    if clip_ds is None:
        raise ValueError(f"Unable to parse input clip geom: {clip_geom}")

    # options
    warp_options = []
    if all_touch:
        warp_options.append("CUTLINE_ALL_TOUCHED=TRUE")
    else:
        warp_options.append("CUTLINE_ALL_TOUCHED=FALSE")

    origin_layer = core_raster.raster_open(raster)

    raster_metadata = core_raster._get_basic_metadata_raster(raster)
    origin_projection = raster_metadata["projection_osr"]

    # Fast check: Does the extent of the two inputs overlap?
    has_inf = True in [np.isinf(val) for val in raster_metadata["bbox_latlng"]]
    if not has_inf and not utils_bbox._check_bboxes_intersect(raster_metadata["bbox_latlng"], clip_metadata["bbox_latlng"]):
        raise ValueError(f"Geometries of {raster} and {clip_geom} did not intersect.")

    if not origin_projection.IsSame(clip_metadata["projection_osr"]):
        clip_ds = _vector_reproject(clip_ds, origin_projection)
        clip_metadata = core_vector._get_basic_metadata_vector(clip_ds)
        memory_files.append(clip_ds)

    output_bounds = raster_metadata["bbox"]

    if crop_to_geom:

        if adjust_bbox:
            output_bounds = utils_bbox._get_aligned_bbox_to_pixel_size(
                raster_metadata["bbox"],
                clip_metadata["bbox"],
                raster_metadata["pixel_width"],
                raster_metadata["pixel_height"],
            )
        else:
            output_bounds = clip_metadata["bbox"]

    # formats
    out_name = path_list[0]
    out_format = utils_gdal._get_raster_driver_name_from_path(out_name)
    out_creation_options = utils_gdal._get_default_creation_options(creation_options)

    # nodata
    if src_nodata == "infer":
        src_nodata = raster_metadata["nodata_value"]
    elif isinstance(src_nodata, (int, float)) or src_nodata is None:
        src_nodata = float(src_nodata)
    else:
        raise ValueError(f"src_nodata must be an int, float or None: {src_nodata}")

    out_nodata = None
    if dst_nodata == "infer":
        if src_nodata == "infer":
            out_nodata = raster_metadata["nodata_value"]
        else:
            out_nodata = utils_translate._get_default_nodata_value(raster_metadata["dtype_gdal"])
    elif isinstance(dst_nodata, (int, float)) or dst_nodata is None:
        out_nodata = dst_nodata
    else:
        raise ValueError(f"Unable to parse nodata_value: {dst_nodata}")

    if out_nodata is not None and not utils_translate._check_is_value_within_dtype_range(out_nodata, raster_metadata["dtype"]):
        warn("Nodata value is outside of the range of the input raster's dtype. Setting to None.")
        out_nodata = None

    # Removes file if it exists and overwrite is True.
    utils_path._delete_if_required(out_path, overwrite)

    if verbose == 0:
        gdal.PushErrorHandler("CPLQuietErrorHandler")

    options = gdal.WarpOptions(
        format=out_format,
        resampleAlg=utils_translate._translate_resample_method(resample_alg),
        targetAlignedPixels=False,
        outputBounds=utils_bbox._get_gdal_bbox_from_ogr_bbox(output_bounds),
        xRes=raster_metadata["pixel_width"],
        yRes=raster_metadata["pixel_height"],
        cutlineDSName=clip_ds,
        cropToCutline=False,
        creationOptions=out_creation_options,
        warpMemoryLimit=utils_gdal._get_dynamic_memory_limit(ram, min_mb=ram_min, max_mb=ram_max),
        warpOptions=warp_options,
        srcNodata=src_nodata,
        dstNodata=out_nodata,
        multithread=True,
    )

    clipped = gdal.Warp(
        out_name,
        origin_layer,
        options=options,
    )

    utils_gdal.delete_dataset_if_in_memory_list(memory_files)

    if verbose == 0:
        gdal.PopErrorHandler()

    if clipped is None:
        raise ValueError("Error while clipping raster.")

    return out_name


def raster_clip(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    clip_geom: Union[str, ogr.DataSource, gdal.Dataset, ogr.Geometry],
    out_path: Optional[str] = None,
    *,
    resample_alg: str = "nearest",
    crop_to_geom: bool = True,
    adjust_bbox: bool = False,
    all_touch: bool = False,
    to_extent: bool = False,
    layer_to_clip: int = 0,
    dst_nodata: Union[float, int, str] = "infer",
    src_nodata: Union[float, int, str] = "infer",
    creation_options: Optional[List[str]] = None,
    add_uuid: bool = False,
    add_timestamp: bool = False,
    prefix: str = "",
    suffix: str = "",
    overwrite: bool = True,
    verbose: int = 0,
    ram: float = 0.8,
    ram_max: Optional[int] = None,
    ram_min: Optional[int] = 100,
):
    """
    Clips a raster(s) using a vector geometry or the extents of a raster.

    Parameters
    ----------
    raster : str or gdal.Dataset or list of str/gdal.Dataset
        The raster(s) to clip.

    clip_geom : str or ogr.DataSource, gdal.Dataset or ogr.Geometry
        The geometry to use to clip the raster.

    out_path : str or list or None, optional
        The path(s) to save the clipped raster to. If None, a memory raster is created. Default: None.

    resample_alg : str, optional
        The resampling algorithm to use. Options include: nearest, bilinear, cubic, cubicspline, lanczos, 
        average, mode, max, min, median, q1, q3, sum, rms. Default: "nearest".

    crop_to_geom : bool, optional
        If True, the output raster will be cropped to the extent of the clip geometry. Default: True.

    adjust_bbox : bool, optional
        If True, the output raster will have its bbox adjusted to match the clip geometry. Default: False.

    all_touch : bool, optional
        If true, all pixels touching the clipping geometry will be included. Default: False.

    to_extent : bool, optional
        If True, the output raster will be cropped to the extent of the clip geometry. Default: False.

    prefix : str, optional
        The prefix to use for the output raster. Default: "".

    suffix : str, optional
        The suffix to use for the output raster. Default: "".

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists. Default: True.

    creation_options : list or None, optional
        A list of creation options to pass to gdal. Default: None.

    dst_nodata : int or float or None, optional
        The nodata value to use for the output raster. Default: "infer".

    src_nodata : int or float or None, optional
        The nodata value to use for the input raster. Default: "infer".

    layer_to_clip : int or str, optional
        The layer ID or name in the vector to use for clipping. Default: 0.

    verbose : int, optional
        The verbosity level. Default: 0.

    add_uuid : bool, optional
        If True, a UUID will be added to the output raster. Default: False.

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output raster. Default: False.

    ram : float, optional
        The proportion of total ram to allow usage of. Default: 0.8.
    
    ram_max: int, optional
        The maximum amount of ram to use in MB. Default: None.
    
    ram_min: int, optional
        The minimum amount of ram to use in MB. Default: 100.

    Returns
    -------
    str or list
        A string or list of strings representing the path(s) to the clipped raster(s).
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(clip_geom, [str, ogr.DataSource, gdal.Dataset, ogr.Geometry], "clip_geom")
    utils_base._type_check(out_path, [[str], str, None], "out_path")
    utils_base._type_check(resample_alg, [str], "resample_alg")
    utils_base._type_check(crop_to_geom, [bool], "crop_to_geom")
    utils_base._type_check(adjust_bbox, [bool], "adjust_bbox")
    utils_base._type_check(all_touch, [bool], "all_touch")
    utils_base._type_check(to_extent, [bool], "to_extent")
    utils_base._type_check(dst_nodata, [str, int, float, None], "dst_nodata")
    utils_base._type_check(src_nodata, [str, int, float, None], "src_nodata")
    utils_base._type_check(layer_to_clip, [int], "layer_to_clip")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "postfix")
    utils_base._type_check(verbose, [int], "verbose")
    utils_base._type_check(add_uuid, [bool], "uuid")
    utils_base._type_check(add_timestamp, [bool], "timestamp")
    utils_base._type_check(ram, [float], "ram")
    utils_base._type_check(ram_max, [int, None], "ram_max")
    utils_base._type_check(ram_min, [int, None], "ram_min")

    input_is_list = isinstance(raster, list)
    input_rasters = utils_io._get_input_paths(raster, "raster")
    out_path_list = utils_io._get_output_paths(
        raster,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_uuid,
        change_ext="tif",
        overwrite=overwrite,
    )
    utils_path._delete_if_required_list(out_path_list, overwrite)

    output = []
    for index, in_raster in enumerate(input_rasters):
        output.append(
            _raster_clip(
                in_raster,
                clip_geom,
                out_path=out_path_list[index],
                resample_alg=resample_alg,
                crop_to_geom=crop_to_geom,
                adjust_bbox=adjust_bbox,
                all_touch=all_touch,
                to_extent=to_extent,
                dst_nodata=dst_nodata,
                src_nodata=src_nodata,
                layer_to_clip=layer_to_clip,
                overwrite=overwrite,
                creation_options=creation_options,
                prefix=prefix,
                suffix=suffix,
                verbose=verbose,
                ram=ram,
                ram_max=ram_max,
                ram_min=ram_min,
            )
        )

    if input_is_list:
        return output

    return output[0]



"""
### Generic utility functions ###

Functions that make interacting with the toolbox easier.
"""

# Standard Library
import sys; sys.path.append("../../")
import os
import shutil
import fnmatch
from glob import glob
from uuid import uuid4
from typing import List, Optional, Union
from warnings import warn

# External
from osgeo import gdal, ogr

# Internal
from buteo.utils import utils_base, utils_gdal



def _get_vsimem_content(
    folder_name: Optional[str] = None,
) -> List[str]:
    """ 
    Get the content of the vsimem folder.
    vsimem is the virtual memory folder of GDAL.
    By default, the content of the root folder is returned.

    Parameters
    ----------
    folder_name: str, optional
        The name of the folder to get the content of. Default: None (root folder)

    Returns
    -------
    List[str]
        A list of the content of the vsimem folder.
    """
    if folder_name is None:
        folder_name = "/vsimem/"

    if not isinstance(folder_name, str):
        raise TypeError("folder_name must be a string.")

    if not folder_name.endswith("/"):
        folder_name += "/"

    try:
        if hasattr(gdal, "listdir"):
            vsimem = gdal.listdir(folder_name)
            vsimem = [folder_name + v.name for v in vsimem]
        elif hasattr(gdal, "ReadDir"):
            vsimem = [folder_name + ds for ds in gdal.ReadDirRecursive(folder_name)]
        else:
            warn("WARNING: Unable to access vsimem. Is GDAL installed?")
            return False

    except RuntimeError as e:
        warn(f"WARNING: Failed to access vsimem. Is GDAL installed? Error: {e}")

    paths = _get_unix_path(vsimem)

    return paths


def _glob_vsimem(
    pattern: str,
) -> List[str]:
    """
    Find files in vsimem using glob.

    Example:
    `_glob_vsimem("*/patches/*.tif")` will return all tif files in the patches folder of all subfolders.

    Parameters
    ----------
    pattern: str
        The pattern to match.

    Returns
    -------
    List[str]
        A list of the files matching the pattern.
    """
    assert isinstance(pattern, str), "pattern must be a string."

    virtual_fs = _get_vsimem_content()
    matches = [path for path in virtual_fs if fnmatch.fnmatch(path, pattern)]

    return matches


def _get_unix_path(path: Union[str, List[str]]) -> Union[str, List[str]]:
    """
    Convert a path or list of paths to a unix path(s).

    Parameters
    ----------
    path: str
        The path to convert.

    Returns
    -------
    Union[str, List[str]]
        The converted path[s].
    """
    input_is_list = False
    if isinstance(path, list):
        input_is_list = True

    path = [path] if isinstance(path, str) else path
    for p in path:
        assert isinstance(p, str), "path must be a string."
        assert len(p) > 0, "path must not be empty."

    out_path = []
    for p in path:
        out_path.append(
            "/".join(os.path.normpath(p).split("\\"))
        )

    if input_is_list:
        return out_path

    return out_path[0]


def _check_file_exists(path: str) -> bool:
    """
    Check if a file exists. Also checks vsimem.
    Handles both absolute and relative paths and with or without trailing slash.

    Parameters
    ----------
    path: str
        The path to the file.

    Returns
    -------
    bool
        True if the file exists, False otherwise.
    """
    if not isinstance(path, str):
        return False

    if os.path.isfile(path):
        return True

    if os.path.isfile(os.path.abspath(path)):
        return True

    vsimem = _get_vsimem_content()
    if _get_unix_path(path) in vsimem:
        return True

    return False


def _check_file_exists_vsimem(path: str) -> bool:
    """
    Check if a file exists in vsimem.

    Parameters
    ----------
    path: str
        The path to the file.

    Returns
    -------
    bool
        True if the file exists, False otherwise.
    """
    if not isinstance(path, str):
        return False

    vsimem = _get_vsimem_content()
    if _get_unix_path(path) in vsimem:
        return True

    return False


def _check_dir_exists(
    path: str,
) -> bool:
    """
    Check if a folder exists. Also checks vsimem.
    Handles both absolute and relative paths and with or without trailing slash.

    Parameters
    ----------
    path: str
        The path to the folder.

    Returns
    -------
    bool
        True if the folder exists, False otherwise.
    """
    if not isinstance(path, str):
        return False

    if os.path.isdir(os.path.normpath(path)):
        return True

    if os.path.isdir(os.path.abspath(path)):
        return True

    split = os.path.normpath(os.path.dirname(path)).split("\\")

    if "vsimem" in split or "vsizip" in split:
        return True

    return False


def _check_dir_exists_vsimem(path: str) -> bool:
    """
    Check if a folder exists in vsimem.

    Parameters
    ----------
    path: str
        The path to the folder.

    Returns
    -------
    bool
        True if the folder exists, False otherwise.
    """
    if not isinstance(path, str):
        return False

    vsimem = _get_vsimem_content()
    if os.path.normpath(path) in vsimem:
        return True

    return False


def _delete_dir_content(
    folder: str,
    delete_subfolders: bool = True,
) -> bool:
    """
    Delete all files and folders in a folder.
    If only the files are to be deleted, set delete_subfolders to False.

    Parameters
    ----------
    folder: str
        The path to the folder.

    delete_subfolders: bool, optional. 
        If True, delete subfolders as well. Default: True

    Returns
    -------
    bool
        True if the folder was deleted, False otherwise.
    """
    assert isinstance(folder, str), "folder must be a string."
    assert len(folder) > 0, "folder must not be a non-empty string."
    assert _check_dir_exists(folder), "folder must exist."

    # Folder is on disk
    if not _check_dir_exists_vsimem(folder):
        for f in os.listdir(folder):
            try:
                path = os.path.join(folder, f)
                if os.path.isfile(path):
                    os.remove(path)

                elif delete_subfolders and os.path.isdir(path):
                    shutil.rmtree(path)

            except RuntimeError:
                warn(f"Warning. Could not remove: {path}", UserWarning)

                return False

    # Folder is in vsimem
    else:
        # Delete the files and folder in VSIMEM
        vsimem = _get_vsimem_content(folder)
        for f in vsimem:
            gdal.Unlink(f)

    if _check_dir_exists(folder):
        warn(f"Warning. Failed to remove: {folder}", UserWarning)
        return False

    return True


def _delete_dir(folder: str) -> bool:
    """
    Delete a folder and all its content. Also deletes vsimem folders.

    Args:
        folder (str): The path to the folder.

    Returns:
        bool: True if the folder was deleted, False otherwise.
    """
    assert isinstance(folder, str), "folder must be a string."
    assert len(folder) > 0, "folder must be a non-empty string."
    assert _check_dir_exists(folder), "folder must exist."

    # Folder is on disk
    if not _check_dir_exists_vsimem(folder):
        try:
            shutil.rmtree(folder)

        except RuntimeError:
            warn(f"Warning. Could not remove: {folder}", UserWarning)

            return False

        return True

    # Delete the files and folder in VSIMEM
    vsimem = _get_vsimem_content(folder)
    for f in vsimem:
        gdal.Unlink(f)

    gdal.Unlink(folder)

    if _check_dir_exists(folder):
        warn(f"Warning. Failed to remove: {folder}", UserWarning)
        return False

    return True


def _delete_file(file: str) -> bool:
    """
    Delete a File

    Args:
        file (str): The path to the file.

    Returns:
        bool: True if the file was deleted, False otherwise.
    """
    assert isinstance(file, str), "file must be a string."
    assert len(file) > 0, "file must be a non-empty string."
    assert _check_file_exists(file), "file must exist."

    # File is on disk
    if not _check_file_exists_vsimem(file):
        try:
            os.remove(file)

        except RuntimeError:
            warn(f"Warning. Could not remove: {file}", UserWarning)

            return False

        return True

    # Delete the file in VSIMEM
    gdal.Unlink(file)

    if _check_file_exists_vsimem(file):
        warn(f"Warning. Failed to remove: {file}", UserWarning)
        return False

    return True


def _create_dir_if_not_exists(path: str) -> str:
    """
    Make a directory if it does not exist.
    This does not work for creating directories in GDAL vsimem.
    On vsimem, it is not possible to create a folder without a file in it.

    If the folder already exists in vsimem, the path to the folder is returned.

    Parameters
    ----------
    path: str
        The path to the folder.

    Returns
    -------
    str
        The path to the folder.
    """
    assert isinstance(path, str), "path must be a string."
    assert len(path) > 0, "path must not be empty."

    if not _check_dir_exists(path):

        # Folder is on disk
        if not _check_dir_exists_vsimem(path):
            os.makedirs(path)

        # Folder is in vsimem
        return path

    if not _check_dir_exists(path):
        raise RuntimeError(f"Could not create folder: {path}.")

    return path


def _get_dir_from_path(path: str) -> str:
    """
    Get the directory of a file. Also works for folders and vsimem.

    Parameters
    ----------
    path: str
        The path to the file.

    Returns
    -------
    str
        The directory of the file.
    """
    assert isinstance(path, str), "path must be a string."
    assert len(path) > 0, "path must not be empty."

    dirs = os.path.normpath(path).split(os.path.sep)[0:2]
    if "vsimem" in dirs:
        dirname = "/vsimem/"
    elif "vsizip" in dirs:
        dirname = "vsizip"
    elif _check_dir_exists(path):
        dirname = _get_unix_path(os.path.abspath(path)) + "/"
    else:
        dirname = _get_unix_path(os.path.dirname(os.path.abspath(path))) + "/"

    return dirname


def _get_filename_from_path(
    path: str,
    with_ext: bool = True,
) -> str:
    """
    Get the filename of a file. Also works for vsimem.

    Parameters
    ----------
    path: str
        The path to the file.

    with_ext: bool
        If True, the extension is included in the filename.

    Returns
    -------
    str
        The filename of the file.
    """
    assert isinstance(path, str), "path must be a string."
    assert len(path) > 0, "path must not be empty."

    basename = os.path.basename(path)

    if with_ext:
        return basename

    basesplit = os.path.splitext(basename)
    return basesplit[0]


def _get_ext_from_path(path: str) -> str:
    """
    Get the extension of a file. If the file has no extension, raise an error.
    The extension is returned without a dot.

    Parameters
    ----------
    path: str
        The path to the file.

    Returns
    -------
    str
        The extension of the file.
    """
    assert isinstance(path, str), "path must be a string."
    assert len(path) > 0, "path must not be empty."

    basename = os.path.basename(path)
    basesplit = os.path.splitext(basename)
    ext = basesplit[1]

    if ext == "" or len(ext) == 1:
        raise RuntimeError (f"File: {path} has no extension.")

    return ext[1:]


def _get_changed_path_ext(
    path: str,
    target_ext: str,
) -> str:
    """
    Update the extension of a file.

    Parameters
    ----------
    path: str
        The path to the file.

    target_ext: str
        The new extension (with or without dot.)
    
    Returns
    -------
    str
        The path to the file with the new extension.
    """
    assert isinstance(path, str), "path must be a string."
    assert isinstance(target_ext, str), "target_ext must be a string."
    assert len(path) > 0, "path must not be empty."
    assert len(target_ext) > 0, "target_ext must not be empty."

    target_ext = target_ext.lstrip('.')
    basename = os.path.basename(path)
    basesplit = os.path.splitext(basename)
    ext = basesplit[1].lstrip('.')

    if ext == "":
        raise RuntimeError(f"File: {path} has no extension.")

    if target_ext == "":
        return os.path.join(os.path.dirname(path), basesplit[0])

    return os.path.join(os.path.dirname(path), f"{basesplit[0]}.{target_ext}")


def _check_is_dir(path: str) -> bool:
    """
    Check if a path is a directory or a file.
    Also returns true if vsimem or vsizip is passed.

    Parameters
    ----------
    path: str
        The path to test.

    Returns
    -------
    bool
        True if path is a directory, False otherwise.
    """
    if not isinstance(path, str):
        return False

    if len(path) == 0:
        return False

    if path in ["/vsimem", "/vsimem/", "/vsizip", "/vsizip/"]:
        return True

    if os.path.isdir(path):
        return True

    return False


def _check_is_valid_mem_filepath(path: str) -> bool:
    """
    Check if a path is a valid memory path that has an extension. vsizip also works.

    Parameters
    ----------
    path: str
        The path to test.

    Returns
    -------
    bool
        True if path is a valid memory path, False otherwise.
    """
    if not isinstance(path, str):
        return False

    if len(path) == 0:
        return False

    path_chunks = os.path.normpath(path).split(os.path.sep)

    if "vsimem" not in path_chunks and "vsizip" not in path_chunks:
        return False

    if _get_ext_from_path(path) == "":
        return False

    return True


def _check_is_valid_non_mem_filepath(path: str) -> bool:
    """
    Check if a path is a valid path that has an extension.
    Does not check if the file exists and does not check if the path is in memory.

    Parameters
    ----------
    path: str
        The path to test.

    Returns
    -------
    bool
        True if path is a valid path, False otherwise.
    """
    if not isinstance(path, str):
        return False

    if len(path) == 0:
        return False

    if _get_ext_from_path(path) == "":
        return False

    folder = _get_dir_from_path(path)
    if not _check_dir_exists(folder):
        return False

    return True


def _check_is_valid_filepath(path: str) -> bool:
    """
    Check if a path is a valid path that has an extension. Path can be in memory.

    Parameters
    ----------
    path: str
        The path to test.

    Returns
    -------
    bool
        True if path is a valid path, False otherwise.
    """
    if _check_is_valid_mem_filepath(path) or _check_is_valid_non_mem_filepath(path):
        return True

    return False


def _check_is_valid_filepath_list(path_list: List[str]) -> bool:
    """
    Check if a list of paths is a valid list of paths that have an extension. Paths can be in memory.

    Parameters
    ----------
    path_list: List[str]
        The list of paths to test.

    Returns
    -------
    bool
        True if path_list is a valid list of paths, False otherwise.
    """
    if not isinstance(path_list, list):
        return False

    if len(path_list) == 0:
        return False

    for path in path_list:
        if not _check_is_valid_filepath(path):
            return False

    return True


def _check_is_valid_output_filepath(
    path: str,
    overwrite: bool = True,
):
    """
    Check if a path is a valid output path that has an extension. Path can be in memory.
    If the file already exists, and overwrite is false, return False.

    Parameters
    ----------
    path: str
        The path to test.
    
    overwrite: bool
        True if the file could be overwritten, False otherwise.

    Returns
    -------
    bool
        True if path is a valid path, False otherwise.
    """
    if not _check_is_valid_filepath(path):
        return False

    if not overwrite:
        if _check_file_exists(path):
            return False

    return True


def _check_is_valid_output_path_list(
    output_list: List[str],
    overwrite: bool = True,
) -> bool:
    """
    Check if a list of output paths are valid.

    Parameters
    ----------
    output_list: list[str]
        The list of paths to the files.

    overwrite: bool
        True if the file should be overwritten, False otherwise.

    Returns
    -------
    bool
        True if the list of output paths are valid, False otherwise.
    """
    if not isinstance(output_list, list):
        return False

    if len(output_list) == 0:
        return False

    for path in output_list:

        if not _check_is_valid_output_filepath(path, overwrite=overwrite):
            return False

    return True


def _delete_if_required(
    path: str,
    overwrite: bool = True,
) -> bool:
    """
    Delete a file if overwrite is True.

    Parameters
    ----------
    path: str
        The path to the file.

    overwrite: bool
        If True, overwrite the file.

    Returns
    -------
    bool
        True if successful, raises error otherwise.
    """
    assert isinstance(path, str), "path must be a string."
    assert len(path) > 0, "path must not be non-empty string."
    assert isinstance(overwrite, bool), "overwrite must be a bool."

    if not overwrite:
        return True

    if not _check_file_exists(path):
        return True

    if _check_is_valid_mem_filepath(path):
        gdal.Unlink(path)
    else:
        try:
            os.remove(path)
        except RuntimeError as e:
            raise RuntimeError(f"Error while deleting file: {path}, {e}") from None

    if _check_file_exists(path):
        raise RuntimeError(f"Error while deleting file: {path}")

    return True


def _delete_if_required_list(
    output_list: List[str],
    overwrite: bool = True,
) -> bool:
    """
    Delete a list of files if overwrite is True.

    Parameters
    ----------
    output_list: list[str]
        The list of paths to the files.

    overwrite: bool
        If True, overwrite the files.

    Returns
    -------
    bool
        True if the files were deleted, False otherwise.
    """
    if not isinstance(output_list, list):
        return False

    if len(output_list) == 0:
        return False

    success = []
    for path in output_list:
        success.append(_delete_if_required(path, overwrite=overwrite))

    if not all(success):
        raise RuntimeError(f"Error while deleting files: {output_list}")

    return True


def _get_changed_folder(
    path: str,
    target_folder: str,
) -> str:
    """
    Change the folder of a path.

    Parameters
    ----------
    path: str
        The path to the file.

    target_folder: str
        The target folder.

    Returns
    -------
    str
        The path with the new folder.
    """
    assert isinstance(path, str), "path must be a string."
    assert len(path) > 0, "path must not be non-empty string."
    assert isinstance(target_folder, str), "target_folder must be a string."
    assert len(target_folder) > 0, "target_folder must not be non-empty string."

    filename = _get_filename_from_path(path)
    joined = os.path.join(target_folder, filename)

    unix_path = _get_unix_path(joined)

    return unix_path


def _check_is_path_glob(path: str) -> bool:
    """
    Check if a path is a glob.

    Parameters
    ----------
    path: str
        The path to check.

    Returns
    -------
    bool
        True if the path is a glob, False otherwise.
    """
    assert isinstance(path, str), "path must be a string."
    assert len(path) > 0, "path must not be non-empty string."

    if path[-5:] == ":glob":
        return True

    return False


def _get_paths_from_glob(path: str) -> List[str]:
    """
    Get a list of paths from a glob.

    Parameters
    ----------
    path: str
        The path to the glob.

    Returns
    -------
    list[str]
        The list of paths.
    """
    assert isinstance(path, str), "path must be a string."
    assert len(path) > 0, "path must not be non-empty string."
    assert _check_is_path_glob(path), "path must be a glob."

    pre_glob = path[:-5]

    if len(pre_glob) > 6 and "vsimem" in pre_glob[:10]:
        return _glob_vsimem(pre_glob)

    return glob(pre_glob)


def _parse_path(path: str) -> str:
    """
    Parse a path to an absolute unix path.

    Parameters
    ----------
    path: str
        The path to parse.

    Returns
    -------
    str
        The parsed path.
    """
    assert isinstance(path, str), "path must be a string."

    abspath = os.path.abspath(path)
    if "\\vsimem\\" in abspath:
        abspath = "/" + abspath.replace(os.path.abspath(os.sep), "")

    abspath = _get_unix_path(abspath)
    return abspath


def _get_augmented_path(
    path: str,
    prefix: str = "",
    suffix: str = "",
    change_ext: Optional[str] = None,
    folder: Optional[str] = None,
    add_uuid: bool = False,
    add_timestamp: bool = False,
) -> str:
    """
    Augments a path with a prefix, suffix, and uuid.
    Can also change the output directory.

    `{prefix}{filename}{uuid}{timestamp}{suffix}.{ext}`

    Parameters
    ----------
    path: str
        The path to the original file.

    prefix: str
        The prefix to add to the path. Default: "".

    suffix: str
        The suffix to add to the path. Default: "".

    change_ext: str. Optional.
        The extension to change the file to. Default: None.

    add_uuid: bool
        If True, add a uuid the path. Default: False.

    add_timestamp: bool
        If True, add a timestamp to the path. Default: False.
        Format: YYYYMMDD_HHMMSS.

    folder: str. Optional.
        The folder to save the file in. This can be specified as
        /vsimem/ to save the file in memory. Default: None.

    Returns
    -------
    str
        The augmented path.
    """
    assert isinstance(path, str), "path must be a string."
    assert isinstance(prefix, str), "prefix must be a string."
    assert isinstance(suffix, str), "suffix must be a string."
    assert isinstance(add_uuid, bool), "add_uuid must be a bool."
    assert folder is None or isinstance(folder, str), "folder must be a string."
    assert change_ext is None or isinstance(change_ext, str), "change_ext must be a string."
    assert len(path) > 0, "path must not be non-empty string."

    path = os.path.abspath(path)
    if "\\vsimem\\" in path:
        path = "/" + path.replace(os.path.abspath(os.sep), "")

    path = _get_unix_path(path)

    # Find the target folder
    target_folder = _get_dir_from_path(path)
    if folder is not None:
        assert len(folder) > 0, "folder must not be non-empty string."
        target_folder = _get_dir_from_path(folder)

    # Find the target extension
    ext = _get_ext_from_path(path)
    if change_ext is not None:
        assert len(change_ext) > 0, "change_ext must not be non-empty string."
        assert isinstance(change_ext, str), "change_ext must be a string."
        ext = change_ext

    ext = ext.lstrip(".").lower()

    filename = _get_filename_from_path(path, with_ext=False)

    if add_uuid:
        uuid = "_" + str(uuid4().int)
    else:
        uuid = ""

    if add_timestamp:
        timestamp = "_" + utils_base._get_time_as_str()
    else:
        timestamp = ""

    augmented_filename = f"{prefix}{filename}{uuid}{timestamp}{suffix}.{ext}"
    augmented_path = os.path.join(target_folder, augmented_filename)

    augmented_path = _get_unix_path(augmented_path)

    return augmented_path


def _get_augmented_path_list(
    path_list: List[str],
    prefix: str = "",
    suffix: str = "",
    change_ext: Optional[str] = None,
    folder: Optional[str] = None,
    add_uuid: bool = False,
    add_timestamp: bool = False,
) -> List[str]:
    """
    Augments a list of paths with a prefix, suffix, and uuid.
    Can also change the output directory.

    Parameters
    ----------
    path_list: list[str]
        The list of paths to the original files.

    prefix: str
        The prefix to add to the path. Default: "".

    suffix: str
        The suffix to add to the path. Default: "".

    change_ext: str. Optional.
        The extension to change the file to. Default: None.

    folder: str. Optional.
        The folder to save the file in. This can be specified as
        /vsimem/ to save the file in memory.

    add_uuid: bool
        If True, add a uuid the path. Default: False.

    Returns
    -------
    list[str]
        The augmented paths.
    """
    assert isinstance(path_list, list), "path_list must be a list."
    assert len(path_list) > 0, "path_list must not be empty."
    assert isinstance(prefix, str), "prefix must be a string."
    assert isinstance(suffix, str), "suffix must be a string."
    assert isinstance(add_uuid, bool), "add_uuid must be a bool."
    assert change_ext is None or isinstance(change_ext, str), "change_ext must be a string."
    assert folder is None or isinstance(folder, str), "folder must be a string."
    assert change_ext is None or isinstance(change_ext, str), "change_ext must be a string."
    assert isinstance(add_timestamp, bool), "add_timestamp must be a bool."

    augmented_path_list = []
    for path in path_list:
        augmented_path_list.append(
            _get_augmented_path(
                path,
                prefix=prefix,
                suffix=suffix,
                change_ext=change_ext,
                folder=folder,
                add_uuid=add_uuid,
                add_timestamp=add_timestamp,
            )
        )

    return augmented_path_list


def _get_temp_filepath(
    name: Union[str, gdal.Dataset, ogr.DataSource] = "temp",
    *,
    ext: Optional[str] = None,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
) -> str:
    """
    Get a temporary filepath in vsimem.

    Parameters
    ----------
    name: Union[str, gdal.Dataset, ogr.DataSource], Optional.
        The name of the file. Default: "temp".

    ext: str, Optional.
        The extension of the file. Default: ".tif".

    prefix: str, Optional.
        The prefix to add to the path. Default: "".

    suffix: str, Optional.
        The suffix to add to the path. Default: "".

    add_uuid: bool, Optional.
        If True, add a uuid the path. Default: True.

    add_timestamp: bool, Optional.
        If True, add a timestamp to the path. Default: True.
        Format: YYYYMMDD_HHMMSS.

    Returns
    -------
    str
        The temporary filepath. (e.g. /vsimem/temp_20210101_000000_123456789.tif)
    """
    assert isinstance(name, (str, ogr.DataSource, gdal.Dataset)), "name must be a string or dataset"
    assert isinstance(ext, (str, type(None))), "ext must be a string or None."
    assert isinstance(prefix, str), "prefix must be a string."
    assert isinstance(suffix, str), "suffix must be a string."
    assert isinstance(add_uuid, bool), "add_uuid must be a bool."
    assert isinstance(add_timestamp, bool), "add_timestamp must be a bool."

    if isinstance(name, gdal.Dataset):
        path = name.GetDescription()
        name = os.path.splitext(os.path.basename(path))[0]
    elif isinstance(name, ogr.DataSource):
        path = name.GetDescription()
        name = os.path.splitext(os.path.basename(path))[0]
    else:
        path = name
        name = os.path.splitext(os.path.basename(name))[0]

    if add_uuid:
        uuid = "_" + str(uuid4().int)
    else:
        uuid = ""

    if add_timestamp:
        timestamp = "_" + utils_base._get_time_as_str()
    else:
        timestamp = ""

    if ext is None:
        ext = _get_ext_from_path(path)

    assert utils_gdal._check_is_valid_ext(ext), f"ext must be a valid extension. {ext} is not valid."

    filename = f"{prefix}{name}{uuid}{timestamp}{suffix}.{ext.lstrip('.').lower()}"
    filepath = os.path.join("/vsimem/", filename)

    # Add _1, _2, etc. if the file already exists
    if _check_file_exists(filepath):
        i = 1
        while _check_file_exists(filepath):
            filename = f"{prefix}{name}{uuid}{timestamp}{suffix}_{i}.{ext.lstrip('.').lower()}"
            filepath = os.path.join("/vsimem/", filename)
            i += 1

    return filepath



"""
### Create grids from rasters.  ###

Cut rasters to grids. Use vectors or rasters as grids.
"""

# TODO: Verify this function and create more grid options.

# Standard library
import sys; sys.path.append("../../")
from uuid import uuid4
from typing import Union, Optional, List

# External
from osgeo import gdal, ogr

# Internal
from buteo.utils import utils_base, utils_gdal, utils_bbox, utils_path
from buteo.raster import core_raster, core_stack
from buteo.raster.clip import _raster_clip
from buteo.vector import core_vector
from buteo.vector.intersect import _vector_intersect
from buteo.vector.reproject import _vector_reproject
from buteo.vector.metadata import _vector_to_metadata


def raster_to_grid(
    raster: Union[str, gdal.Dataset],
    grid: Union[str, ogr.DataSource],
    out_dir: str,
    *,
    use_field: bool = None,
    generate_vrt: bool = True,
    overwrite: bool = True,
    process_layer: int = 0,
    creation_options: Optional[List[str]] = None,
    verbose: int = 0,
) -> str:
    """
    Clips a raster to a grid. Generates .vrt.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset]
        The input raster.

    grid : Union[str, ogr.DataSource]
        The grid to use.

    out_dir : str
        The output directory.

    use_field : bool, optional
        A field to use to name the grid cells, default: None.

    generate_vrt : bool, optional
        If True, the output raster will be a .vrt, default: True.

    overwrite : bool, optional
        If True, the output raster will be overwritten, default: True.

    process_layer : int, optional
        The layer from the grid to process, default: 0.

    creation_options : Optional[List[str]], optional
        Creation options for the output raster, default: None.

    verbose : int, optional
        The verbosity level, default: 0.

    Returns
    -------
    str
        The file path for the newly created raster.
    """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")
    utils_base._type_check(grid, [str, ogr.DataSource], "grid")
    utils_base._type_check(out_dir, [str], "out_dir")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(process_layer, [int], "process_layer")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(verbose, [int], "verbose")

    use_grid = core_vector.vector_open(grid)
    grid_metadata = _vector_to_metadata(use_grid)
    raster_metadata = core_raster._get_basic_metadata_raster(raster)

    # Reproject raster if necessary.
    if not raster_metadata["projection_osr"].IsSame(grid_metadata["projection_osr"]):
        use_grid = _vector_reproject(grid, raster_metadata["projection_osr"])
        grid_metadata = _vector_to_metadata(use_grid)

        if not isinstance(grid_metadata, dict):
            raise RuntimeError("Error while parsing metadata.")

    # Only use the polygons in the grid that intersect the extent of the raster.
    use_grid = _vector_intersect(use_grid, raster_metadata["extent_datasource"]())

    ref = core_raster._raster_open(raster)
    use_grid = core_vector.vector_open(use_grid)

    layer = use_grid.GetLayer(process_layer)
    feature_count = layer.GetFeatureCount()
    raster_extent = raster_metadata["bbox"]
    filetype = utils_path._get_ext_from_path(raster)
    name = raster_metadata["name"]
    geom_type = grid_metadata["layers"][process_layer]["geom_type_ogr"]

    if use_field is not None:
        if use_field not in grid_metadata["layers"][process_layer]["field_names"]:
            names = grid_metadata["layers"][process_layer]["field_names"]
            raise ValueError(
                f"Requested field not found. Fields available are: {names}"
            )

    generated = []

    # For the sake of good reporting - lets first establish how many features intersect
    # the raster.

    if verbose:
        print("Finding intersections.")

    intersections = 0
    for _ in range(feature_count):
        feature = layer.GetNextFeature()
        geom = feature.GetGeometryRef()

        if not utils_bbox._check_bboxes_intersect(raster_extent, geom.GetEnvelope()):
            continue

        intersections += 1

    layer.ResetReading()

    if verbose:
        print(f"Found {intersections} intersections.")

    if intersections == 0:
        print("Warning: Found 0 intersections. Returning empty list.")
        return ([], None)

    driver = ogr.GetDriverByName("GPKG")

    clipped = 0
    for _ in range(feature_count):

        feature = layer.GetNextFeature()
        geom = feature.GetGeometryRef()

        if not utils_bbox._check_bboxes_intersect(raster_extent, geom.GetEnvelope()):
            continue

        if verbose == 1:
            utils_base.progress(clipped, intersections - 1, "clip_grid")

        fid = feature.GetFID()

        test_ds_path = f"/vsimem/grid_{uuid4().int}.gpkg"
        test_ds = driver.CreateDataSource(test_ds_path)
        test_ds_lyr = test_ds.CreateLayer(
            "mem_layer_grid",
            geom_type=geom_type,
            srs=raster_metadata["projection_osr"],
        )
        test_ds_lyr.CreateFeature(feature.Clone())
        test_ds.FlushCache()

        out_name = None

        if use_field is not None:
            out_name = f"{out_dir}{feature.GetField(use_field)}{filetype}"
        else:
            out_name = f"{out_dir}{name}_{fid}{filetype}"

        _raster_clip(
            ref,
            test_ds_path,
            out_path=out_name,
            adjust_bbox=True,
            crop_to_geom=True,
            all_touch=False,
            suffix="",
            prefix="",
            creation_options=utils_gdal._get_default_creation_options(creation_options),
            verbose=0,
        )

        generated.append(out_name)
        clipped += 1

    if generate_vrt:
        vrt_name = f"{out_dir}{name}.vrt"

        core_stack.raster_stack_vrt_list(generated, vrt_name, separate=False)

        return (generated, vrt_name)

    return generated



"""
### Shift rasters. ###

Module to shift the location of rasters in geographic coordinates.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import gdal
import numpy as np

# Internal
from buteo.utils import (
    utils_base,
    utils_gdal,
    utils_path,
    utils_io,
)
from buteo.raster import core_raster, core_raster_io
from buteo.array.convolution import convolve_array_simple
from buteo.array.convolution_kernels import kernel_shift


def _raster_shift(
    raster: Union[str, gdal.Dataset],
    shift_list: List[Union[int, float]],
    out_path: Optional[str] = None,
    *,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
) -> Union[str, gdal.Dataset]:
    """ Internal. """
    assert isinstance(shift_list, (list, tuple)), f"shift_list must be a list or a tuple. {shift_list}"
    assert len(shift_list) == 2, f"shift_list must be a list or tuple with len 2 (x_shift, y_shift): {shift_list}"

    for shift in shift_list:
        assert isinstance(shift, (int, float)), f"shift must be an int or a float: {shift}"

    ref = core_raster._raster_open(raster)
    metadata = core_raster._get_basic_metadata_raster(ref)

    x_shift, y_shift = shift_list

    if out_path is None:
        out_path = utils_path._get_temp_filepath("shifted_raster.tif")
    else:
        if not utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite):
            raise ValueError(f"out_path is not a valid output path: {out_path}")

    utils_path._delete_if_required(out_path, overwrite)

    driver = gdal.GetDriverByName(utils_gdal._get_raster_driver_name_from_path(out_path))

    shifted = driver.Create(
        out_path,  # Location of the saved raster, ignored if driver is memory.
        metadata["width"],  # Dataframe width in pixels (e.g. 1920px).
        metadata["height"],  # Dataframe height in pixels (e.g. 1280px).
        metadata["bands"],  # The number of bands required.
        metadata["dtype_gdal"],  # Datatype of the destination.
        utils_gdal._get_default_creation_options(creation_options),
    )

    new_transform = list(metadata["geotransform"])
    new_transform[0] += x_shift
    new_transform[3] += y_shift

    shifted.SetGeoTransform(new_transform)
    shifted.SetProjection(metadata["projection_wkt"])

    src_nodata = metadata["nodata_value"]

    for band in range(metadata["bands"]):
        origin_raster_band = ref.GetRasterBand(band + 1)
        target_raster_band = shifted.GetRasterBand(band + 1)

        target_raster_band.WriteArray(origin_raster_band.ReadAsArray())

        if src_nodata is not None:
            if utils_base._check_variable_is_int(src_nodata):
                target_raster_band.SetNoDataValue(int(src_nodata))
            else:
                target_raster_band.SetNoDataValue(float(src_nodata))

    shifted.FlushCache()
    shifted = None
    ref = None

    return out_path


def raster_shift(
    raster: Union[str, gdal.Dataset, List],
    shift_list: List[Union[int, float]],
    out_path: Optional[str] = None,
    *,
    overwrite: bool = True,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    creation_options: Optional[List[str]] = None,
) -> Union[str, List[str], gdal.Dataset]:
    """
    Shifts a raster in a given direction. (The frame is shifted)

    Parameters
    ----------
    raster : Union[str, List, gdal.Dataset]
        The raster(s) to be shifted.

    shift_list : List[Union[int, float]]
        The shift in x and y direction.

    out_path : Optional[str], optional
        The path to the output raster. If None, the raster is created in memory., default: None

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists., default: True

    prefix : str, optional
        The prefix to be added to the output raster name., default: ""

    suffix : str, optional
        The suffix to be added to the output raster name., default: ""

    add_uuid : bool, optional
        If True, a unique identifier will be added to the output raster name., default: False

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output raster name., default: False

    creation_options : Optional[List[str]], optional
        The creation options to be used when creating the output., default: None

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the shifted raster(s).
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(shift_list, [[int, float], tuple], "shift_list")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")

    input_is_list = isinstance(raster, list)

    input_list = utils_io._get_input_paths(raster, "raster")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        change_ext="tif",
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    shifted_rasters = []
    for idx, in_raster in enumerate(input_list):
        shifted_rasters.append(
            _raster_shift(
                in_raster,
                shift_list,
                out_path=output_list[idx],
                overwrite=overwrite,
                creation_options=creation_options,
            )
        )

    if input_is_list:
        return shifted_rasters

    return shifted_rasters[0]


def raster_shift_pixel(
    raster: Union[str, gdal.Dataset],
    shift_list: List[Union[int, float]],
    out_path: Optional[str] = None,
) -> str:
    """
    Shifts a raster in a given direction. (The pixels are shifted, not the frame)

    Parameters
    ----------
    raster : Union[str, gdal.Dataset]
        The raster to be shifted.

    shift_list : List[Union[int, float]]
        The shift in x and y direction.

    out_path : Optional[str], optional
        The path to the output raster. If None, the raster is created in memory., default: None

    Returns
    -------
    str
        The path to the shifted raster.
    """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")
    utils_base._type_check(shift_list, [[int, float], tuple], "shift_list")
    utils_base._type_check(out_path, [str, None], "out_path")

    arr = core_raster_io.raster_to_array(raster)

    offsets, weights = kernel_shift(shift_list[0], shift_list[1])

    arr_float32 = arr.astype(np.float32, copy=False)

    for channel in range(arr.shape[2]):
        arr[:, :, channel] = convolve_array_simple(
            arr_float32[:, :, channel], offsets, weights,
        )

    if out_path is None:
        out_path = utils_path._get_temp_filepath("shifted_raster.tif")
    else:
        if not utils_path._check_is_valid_output_filepath(out_path):
            raise ValueError(f"out_path is not a valid output path: {out_path}")

    utils_path._delete_if_required(out_path, overwrite=True)

    return core_raster_io.array_to_raster(
        arr,
        reference=raster,
        out_path=out_path,
    )



"""
### Clip vectors to other geometries ###

Clip vector files with other geometries. Can come from rasters or vectors.
"""

# Standard library
import sys; sys.path.append("../../")
import os
from typing import Union, Optional, List

# External
from osgeo import ogr

# Internal
from buteo.utils import utils_base, utils_path, utils_io, utils_gdal
from buteo.vector import core_vector


def _vector_buffer(
    vector: Union[str, ogr.DataSource],
    distance: Union[int, float, str],
    out_path: Optional[str] = None,
    in_place: bool = False,
    force_multipolygon: bool = True,
) -> str:
    """ Internal. """
    assert isinstance(vector, (str, ogr.DataSource)), "Invalid vector input."
    assert isinstance(distance, (int, float, str)), "Invalid distance input."
    assert isinstance(out_path, (str, type(None))), "Invalid output path input."

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector, suffix="_buffered", ext="gpkg")
    else:
        assert utils_path._check_is_valid_output_filepath(vector, out_path), "Invalid vector output path."

    read = core_vector.vector_open(vector)
    metadata = core_vector._get_basic_metadata_vector(read)

    if in_place:
        for l in metadata["layers"]:
            if l["geom_type"] not in ["Polygon", "MultiPolygon", "3D Polygon", "3D MultiPolygon"]:
                raise ValueError("Input vector must be polygonal to be buffered in-place.")
    else:
        driver = ogr.GetDriverByName(utils_gdal._get_driver_name_from_path(out_path))
        if driver is None:
            raise Exception("Driver for output buffer file is not available.")
        
        if os.path.exists(out_path):
            driver.DeleteDataSource(out_path)

        output_ds = driver.CreateDataSource(out_path)
        if output_ds is None:
            raise Exception(f"Could not create output file: {out_path}")

    if isinstance(distance, str):
        for layer_meta in metadata["layers"]:
            if distance not in layer_meta["field_names"]:
                raise AttributeError(f"Attribute ({distance}) is a string and not one of the vector field names.")

    for layer_meta in metadata["layers"]:
        layer_name = layer_meta["layer_name"]

        vector_layer_origin = read.GetLayer(layer_name)
        if in_place:
            vector_layer_destination = vector_layer_origin
        else:
            vector_layer_destination = output_ds.CreateLayer(layer_name, geom_type=ogr.wkbMultiPolygon, srs=vector_layer_origin.GetSpatialRef())
            vector_layer_origin_defn = vector_layer_origin.GetLayerDefn()

            for i in range(vector_layer_origin_defn.GetFieldCount()):
                vector_layer_destination.CreateField(vector_layer_origin_defn.GetFieldDefn(i))

        if isinstance(distance, str):
            if distance not in layer_meta["field_names"]:
                raise AttributeError(f"Attribute ({distance}) is a stirng and not one of the vector field names.")
            buffer_is_field = True
        else:
            buffer_is_field = False

        vector_layer_origin = read.GetLayer(layer_name)

        feature_count = vector_layer_origin.GetFeatureCount()

        vector_layer_origin.ResetReading()
        for _ in range(feature_count):
            feature = vector_layer_origin.GetNextFeature()

            if buffer_is_field:
                buffer_distance = feature.GetField(distance)
            else:
                buffer_distance = distance

            if feature is None:
                break

            feature_geom = feature.GetGeometryRef()

            if feature_geom is None:
                continue

            buffered_geometry = feature_geom.Buffer(buffer_distance)

            if force_multipolygon:
                if buffered_geometry.GetGeometryType() == ogr.wkbPolygon:
                    buffered_geometry = ogr.ForceToMultiPolygon(buffered_geometry)

            if in_place:
                try:
                    feature.SetGeometry(buffered_geometry)
                    vector_layer_origin.SetFeature(feature)
                except ValueError:
                    continue

            else:
                out_feature = ogr.Feature(vector_layer_destination.GetLayerDefn())
                out_feature.SetGeometry(buffered_geometry)

                for i in range(feature.GetFieldCount()):
                    out_feature.SetField(i, feature.GetField(i))
                
                vector_layer_destination.CreateFeature(out_feature)

        vector_layer_destination.GetExtent()
        vector_layer_destination.ResetReading()
        vector_layer_destination.SyncToDisk()

    read.FlushCache()
    read = None

    return out_path


def vector_buffer(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    distance: Union[int, float, str],
    out_path: Optional[Union[str, List[str]]] = None,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    in_place: bool = False,
    overwrite: bool = True,
) -> Union[str, List[str]]:
    """
    Buffers a vector with a fixed distance or an attribute.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[str, ogr.DataSource]]
        Vector(s) to buffer.

    distance : Union[int, float, str]
        The distance to buffer with. If string, uses the attribute of that name.

    out_path : Optional[str], optional
        Output path. If None, memory vectors are created. Default: None

    prefix : str, optional
        Prefix to add to the output path. Default: ""

    suffix : str, optional
        Suffix to add to the output path. Default: ""

    add_uuid : bool, optional
        Add UUID to the output path. Default: False

    add_timestamp : bool, optional
        Add timestamp to the output path. Default: False
    
    in_place : bool, optional
        If True, overwrites the input vector. Default: False

    overwrite : bool, optional
        Overwrite output if it already exists. Default: True

    Returns
    -------
    Union[str, List[str]]
        Output path(s) of clipped vector(s).
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    utils_base._type_check(distance, [int, float, str], "distance")
    utils_base._type_check(out_path, [str, None], "out_path")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")
    utils_base._type_check(in_place, [bool], "in_place")
    utils_base._type_check(overwrite, [bool], "overwrite")

    input_is_list = isinstance(vector, list)
    input_data = utils_io._get_input_paths(vector, "vector")
    output_data = utils_io._get_output_paths(
        input_data,
        out_path,
        in_place=in_place,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        overwrite=overwrite,
    )

    if not in_place:
        utils_path._delete_if_required_list(output_data, overwrite)

    output = []
    for idx, in_vector in enumerate(input_data):
        output.append(
            _vector_buffer(
                in_vector,
                distance,
                out_path=output_data[idx],
                in_place=in_place,
            )
        )

    if input_is_list:
        return output

    return output[0]



"""
### Functions to turn rasters into rasters of coordinates. ###
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Tuple

# External
import numpy as np
from osgeo import gdal

# Internal
from buteo.utils import utils_base
from buteo.raster import core_raster


# TODO: Define how this is going to work.
def raster_create_grid_with_coordinates(
    raster: Union[str, gdal.Dataset],
    latlng: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Create a grid of coordinates from a raster. Format is (x, y, xy).

    Parameters
    ----------
    raster : str or gdal.Dataset
        The raster to create the grid from.

    Returns
    -------
    np.ndarray
        A NumPy array of shape (x, y, xy-coordinates).
    """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")

    meta = core_raster._get_basic_metadata_raster(raster)

    step_x = meta["pixel_width"]
    size_x = meta["width"]
    start_x = meta["x_min"]
    stop_x = meta["x_max"]

    step_y = -meta["pixel_height"]
    size_y = meta["height"]
    start_y = meta["y_max"]
    stop_y = meta["y_min"]

    x_adj = step_x / 2
    y_adj = step_y / 2

    x_vals = np.linspace(start_x + x_adj, stop_x - x_adj, size_x, dtype=np.float32)
    y_vals = np.linspace(start_y - y_adj, stop_y + y_adj, size_y, dtype=np.float32)

    xx, yy = np.meshgrid(x_vals, y_vals)
    grid = np.dstack((xx, yy))

    return grid



"""
### Align rasters ###

Functions to align a series of rasters to a master or a reference.
"""

# TODO: phase_cross_correlation

# Standard library
import sys; sys.path.append("../../")
from typing import List, Union, Optional

# External
from osgeo import gdal

# Internal
from buteo.utils import (
    utils_io,
    utils_path,
    utils_gdal,
    utils_base,
    utils_bbox,
    utils_translate,
    utils_projection,
)

from buteo.raster import core_raster, core_raster_io
from buteo.raster.reproject import raster_reproject


def _raster_align_to_reference(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    reference: Union[str, gdal.Dataset],
    *,
    out_path: Optional[Union[str, List[str]]] = None,
    resample_alg: str = "nearest",
    target_nodata: Optional[Union[int, float]] = None,
    overwrite=True,
    creation_options: Optional[List[str]] = None,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    ram: float = 0.8,
    ram_max: Optional[int] = None,
    ram_min: Optional[int] = 100,
) -> List[str]:
    """
    Aligns a series of rasters to a reference.

    Parameters
    ----------
    rasters : list of str
        A list of rasters to align.

    reference : str or gdal.Dataset
        Path to the reference raster or vector.

    out_path : str or list of str, optional
        Paths to the output. If not provided, the output will be in-memory rasters.

    resample_alg : str, optional
        Resampling algorithm to use. Default: "nearest".

    target_nodata : int or float, optional
        Nodata value to use for the output rasters.

    overwrite : bool, optional
        Overwrite existing files. Default: True.

    creation_options : list, optional
        List of creation options.

    prefix : str, optional
        Prefix to add to the output file name. Default: "".

    suffix : str, optional
        Suffix to add to the output file name. Default: "".

    add_uuid : bool, optional
        Whether to add a uuid to the output file name. Default: False.

    add_timestamp : bool, optional
        Whether to add a timestamp to the output file name. Default: False.

    ram : float, optional
        The proportion of total ram to allow usage of. Default: 0.8.
    
    ram_max: int, optional
        The maximum amount of ram to use in MB. Default: None.
    
    ram_min: int, optional
        The minimum amount of ram to use in MB. Default: 100.

    Returns
    -------
    List[str]
        A list of paths to the aligned rasters.
    """
    utils_base._type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")
    utils_base._type_check(reference, [str, gdal.Dataset], "reference")
    utils_base._type_check(out_path, [str, None, [str]], "out_path")
    utils_base._type_check(resample_alg, [str], "resample_alg")
    utils_base._type_check(target_nodata, [int, float, None], "target_nodata")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")
    utils_base._type_check(ram, [float], "ram")
    utils_base._type_check(ram_max, [int, None], "ram_max")
    utils_base._type_check(ram_min, [int, None], "ram_min")

    input_rasters = utils_io._get_input_paths(rasters, "raster")

    assert utils_gdal._check_is_raster_or_vector(reference), "Reference must be a raster or vector."

    # Verify that all of the rasters overlap the reference
    for raster in input_rasters:
        assert core_raster.check_rasters_intersect(raster, reference), (
            f"Raster: {utils_gdal._get_path_from_dataset(raster)} does not intersect reference."
        )

    path_list = utils_io._get_output_paths(
        input_rasters,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        overwrite=overwrite,
        change_ext="tif",
    )

    creation_options = utils_gdal._get_default_creation_options(creation_options)

    reference_metadata = core_raster._get_basic_metadata_raster(reference)

    target_projection = reference_metadata["projection_osr"]
    reference_bbox = reference_metadata["bbox"]
    x_min, _, _, y_max = reference_bbox

    # Set the target values.
    x_res = reference_metadata["pixel_width"]
    y_res = reference_metadata["pixel_height"]
    x_pixels = reference_metadata["width"]
    y_pixels = reference_metadata["height"]

    # Reproject the rasters to the reference projection.
    for idx, raster in enumerate(input_rasters):
        raster_metadata = core_raster._get_basic_metadata_raster(raster)
        raster_path = utils_gdal._get_path_from_dataset(raster)

        raster_reprojected = None
        raster_projection = raster_metadata["projection_osr"]
        if not utils_projection._check_projections_match(raster_projection, target_projection):
            raster_reprojected = raster_reproject(
                raster_path,
                target_projection,
                out_path=None,
                resample_alg=resample_alg,
                creation_options=creation_options,
                prefix="tmp_reprojection_",
                add_uuid=True,
            )
            raster_ds = core_raster._raster_open(raster_reprojected)
        else:
            raster_ds = core_raster._raster_open(raster_path)

        destination_ds = core_raster_io.raster_create_empty(
            out_path=path_list[idx],
            width=x_pixels,
            height=y_pixels,
            pixel_size=(x_res, y_res),
            x_min=x_min,
            y_max=y_max,
            dtype=raster_metadata["dtype"],
            bands=raster_metadata["bands"],
            nodata_value=target_nodata,
            projection=target_projection,
            creation_options=creation_options,
            overwrite=overwrite,
        )

        destination_ds = core_raster._raster_open(destination_ds)

        warp_options = gdal.WarpOptions(
            resampleAlg=utils_translate._translate_resample_method(resample_alg),
            multithread=True,
            outputBounds=utils_bbox._get_gdal_bbox_from_ogr_bbox(reference_bbox),
            warpMemoryLimit=utils_gdal._get_dynamic_memory_limit(ram, min_mb=ram_min, max_mb=ram_max),
        )

        warped = gdal.Warp(destination_ds, raster_ds, options=warp_options)

        if raster_reprojected is not None:
            utils_gdal.delete_dataset_if_in_memory(raster_reprojected)

        destination_ds.FlushCache()
        destination_ds = None

        if warped is None:
            raise ValueError("Error while warping rasters.")

    return path_list


def _raster_find_best_align_reference(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    method: str,
    out_path: Optional[str] = None,
) -> str:
    """
    Find the best reference raster for aligning a list of rasters.

    Parameters
    ----------
    rasters : Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]]
        List of rasters to align.

    method : str
        Bounding box method to use for finding the best reference.
        Options include: "reference", "intersection", and "union".

    out_path : str, optional
        Path to the output raster, default: None

    Returns
    -------
    str
        Path to the best reference raster.
    """
    utils_base._type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")
    utils_base._type_check(method, [str], "method")
    utils_base._type_check(out_path, [str, None], "out_path")

    assert method in ["reference", "intersection", "union"], (
        f"Invalid bounding_box: {method}. Options include: reference, intersection, and union."
    )

    input_rasters = utils_io._get_input_paths(rasters, "raster")

    if len(input_rasters) == 1:
        return input_rasters[0]

    if out_path is None:
        out_path = utils_path._get_temp_filepath(name="best_reference.tif")
    else:
        if not utils_path._check_is_valid_output_filepath(out_path, overwrite=True):
            raise ValueError(f"Invalid out_path: {out_path}")

    # Count intersections
    most_intersections = 0
    intersections_arr = []
    for idx_i, raster_i in enumerate(input_rasters):
        intersections = 0
        for idx_j, raster_j in enumerate(input_rasters):
            if idx_i == idx_j:
                continue

            # Uses the latlng bounding box to determine if the rasters intersect.
            intersection = core_raster.check_rasters_intersect(raster_i, raster_j)

            if intersection:
                intersections += 1

        if intersections > most_intersections:
            most_intersections = intersections

        intersections_arr.append(intersections)

    # Find the raster with the most intersections.
    largest_area = 0.0
    largest_area_idx = 0
    for idx, intersection in enumerate(intersections_arr):
        if intersection == most_intersections:
            raster = input_rasters[idx]
            raster_area = core_raster._get_basic_metadata_raster(raster)["area_latlng"]

            if raster_area > largest_area:
                largest_area = raster_area
                largest_area_idx = idx

    best_reference = utils_gdal._get_path_from_dataset(input_rasters[largest_area_idx])

    input_path_list = []
    for raster in input_rasters:
        input_path_list.append(utils_gdal._get_path_from_dataset(raster))

    all_other_rasters = []
    for raster_path in input_path_list:
        if raster_path == best_reference:
            continue

        all_other_rasters.append(raster_path)

    for raster in all_other_rasters:
        assert core_raster.check_rasters_intersect(best_reference, raster), (
            f"Rasters {best_reference} and {raster} do not intersect."
        )

    if method == "reference":
        return best_reference

    matched_rasters = raster_reproject(
        all_other_rasters,
        best_reference,
        copy_if_same=False,
    )

    best_reference_meta = core_raster._get_basic_metadata_raster(best_reference)

    best_geom = best_reference_meta["geom"]
    for raster_path in matched_rasters:
        raster_meta = core_raster._get_basic_metadata_raster(raster_path)
        raster_geom = raster_meta["geom"]

        if not raster_geom.Intersects(best_geom):
            raise ValueError(
                f"Raster {raster_path} did not intersect. Consider not using bbox='intersection'"
            )

        if method == "union":
            best_geom = raster_geom.Union(best_geom)
        else:
            best_geom = raster_geom.Intersection(best_geom)

    bbox_best_geom = best_geom.GetEnvelope()

    for raster_path in matched_rasters:
        if raster_path not in input_path_list:
            utils_gdal.delete_dataset_if_in_memory(raster_path)

    outbox_bbox = utils_bbox._get_aligned_bbox_to_pixel_size(
        best_reference_meta["bbox"],
        bbox_best_geom,
        best_reference_meta["pixel_width"],
        best_reference_meta["pixel_height"],
    )

    creation_options = utils_gdal._get_default_creation_options()

    width = int(round((outbox_bbox[1] - outbox_bbox[0]) / best_reference_meta["pixel_width"]))
    height = int(round((outbox_bbox[3] - outbox_bbox[2]) / best_reference_meta["pixel_height"]))

    out_path = core_raster_io.raster_create_empty(
        out_path=out_path,
        width=width,
        height=height,
        pixel_width=best_reference_meta["pixel_width"],
        pixel_height=best_reference_meta["pixel_height"],
        bands=1,
        dtype=best_reference_meta["dtype"],
        x_min=outbox_bbox[0],
        y_max=outbox_bbox[3],
        projection=best_reference_meta["projection"],
        creation_options=creation_options,
    )

    return out_path


def raster_align(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    *,
    out_path: Optional[Union[str, List[str]]] = None,
    reference: Optional[Union[str, gdal.Dataset]] = None,
    method: str = "reference",
    target_nodata: Optional[Union[int, float]] = None,
    resample_alg: str = "nearest",
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    ram: float = 0.8,
    ram_min: Optional[int] = 100,
    ram_max: Optional[int] = None,
) -> List[str]:
    """
    Aligns rasters either to a reference raster or to each other using one of three methods:
    reference, intersection, or union.

    Parameters
    ----------
    rasters : Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]]
        The rasters to align.

    out_path : List[str] or str, optional
        The output path(s), default: None

    reference : Str or gdal.Dataset, optional
        The reference raster to align to, default: None

    method : str, optional
        The method to use, default: "reference" [reference, intersection, union]

    resample_alg : str, optional
        The resampling algorithm to use, default: "nearest"

    overwrite : bool, optional
        Whether to overwrite existing files, default: True

    creation_options : Optional[List[str]], optional
        The creation options to use, default: None

    prefix : str, optional
        The prefix to add to the output file name, default: ""

    suffix : str, optional
        The suffix to add to the output file name, default: ""

    add_uuid : bool, optional
        Whether to add a uuid to the output file name, default: False

    add_timestamp : bool, optional
        Whether to add a timestamp to the output file name, default: False

    ram : float, optional
        The proportion of total ram to allow usage of, default: 0.8

    ram_min : int, optional
        The minimum amount of RAM to use in MB, default: 100

    ram_max : int, optional
        The maximum amount of RAM to use in MB, default: None
    
    Returns
    -------
    List[str]
        The aligned rasters.
    """
    utils_base._type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")
    utils_base._type_check(out_path, [str, None, [str]], "out_path")
    utils_base._type_check(reference, [str, [str], None], "reference")
    utils_base._type_check(method, [str], "method")
    utils_base._type_check(target_nodata, [int, float, None], "target_nodata")
    utils_base._type_check(resample_alg, [str], "resample_alg")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")
    utils_base._type_check(ram, [float], "ram")
    utils_base._type_check(ram_min, [int, None], "ram_min")
    utils_base._type_check(ram_max, [int, None], "ram_max")
    assert method in ["reference", "intersection", "union"], "method must be one of reference, intersection, or union."

    raster_list = utils_io._get_input_paths(rasters, "raster")

    path_list = utils_io._get_output_paths(
        raster_list,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        overwrite=overwrite,
        change_ext="tif",
    )

    if reference is None:
        reference = _raster_find_best_align_reference(raster_list, method)

    aligned = _raster_align_to_reference(
        raster_list,
        reference,
        out_path=path_list,
        resample_alg=resample_alg,
        target_nodata=target_nodata,
        overwrite=overwrite,
        creation_options=creation_options,
        suffix=suffix,
        prefix=prefix,
        ram=ram,
        ram_min=ram_min,
        ram_max=ram_max,
    )

    return aligned



"""
This module contains function to select features from a dataset for machine learning.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Optional, Union, Tuple

# External
import numpy as np


def split_train_val(
    X: np.ndarray,
    y: np.ndarray,
    val_size: float = 0.2,
    random_state: Optional[Union[float, int]] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Split a dataset into training and validation sets using random shuffling.

    Parameters
    ----------
    X : np.ndarray
        The input data.

    y : np.ndarray
        The target data.

    val_size : float, optional
        The size of the validation set as a fraction of the total dataset. Default: 0.2.

    random_state : Union[float, int], optional
        The random seed to use for the split. Default is None.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        A tuple containing the training and validation sets for input data and target data,
        respectively, in the following order: x_train, x_val, y_train, y_val.

    Raises
    ------
    AssertionError
        If X and y are not numpy arrays, X and y do not have the same number of rows, or
        val_size is not between 0 and 1.

    Notes
    -----
    The function always shuffles the data before splitting.
    """
    assert isinstance(X, np.ndarray), "X should be a numpy array."
    assert isinstance(y, np.ndarray), "y should be a numpy array."
    assert X.shape[0] == y.shape[0], "X and y should be the same axis=0 size."
    assert 0 < val_size < 1, "val_size should be between 0 and 1."

    if random_state is not None:
        np.random.seed(random_state)

    indices = np.arange(X.shape[0])
    np.random.shuffle(indices)

    split_idx = int(X.shape[0] * (1 - val_size))

    X_train = X[indices[:split_idx]]
    X_test = X[indices[split_idx:]]
    y_train = y[indices[:split_idx]]
    y_test = y[indices[split_idx:]]

    return X_train, X_test, y_train, y_test


def split_train_val_test(
    X: np.ndarray,
    y: np.ndarray,
    val_size: float = 0.1,
    test_size: float = 0.2,
    random_state: Optional[Union[float, int]] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Split a dataset into training, validation, and test sets using a specified random seed.

    Parameters
    ----------
    X : np.ndarray
        The input data to be split.

    y : np.ndarray
        The target data to be split.

    val_size : float, optional
        The proportion of the data to use for validation, default: 0.1.

    test_size : float, optional
        The proportion of the data to use for testing, default: 0.2.

    random_state : float or int, optional
        Seed for the random number generator used for shuffling, default: None.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        A tuple containing the following arrays:
        - x_train: training input data
        - x_val: validation input data
        - x_test: testing input data
        - y_train: training target data
        - y_val: validation target data
        - y_test: testing target data

    Raises
    ------
    AssertionError
        If X and y are not numpy arrays with the same shape[0], or if val_size or test_size are not between 0 and 1.

    Notes
    -----
    The function always shuffles the data before splitting.
    """
    assert isinstance(X, np.ndarray), "X should be a numpy array."
    assert isinstance(y, np.ndarray), "y should be a numpy array."
    assert X.shape[0] == y.shape[0], "X and y should be the same axis=0 size."
    assert 0 < val_size < 1, "val_size should be between 0 and 1."
    assert 0 < test_size < 1, "test_size should be between 0 and 1."

    if random_state is not None:
        np.random.seed(random_state)

    indices = np.arange(X.shape[0])
    np.random.shuffle(indices)

    test_split_idx = int(X.shape[0] * (1 - test_size))
    val_split_idx = int((1 - val_size) * test_split_idx)

    X_train = X[indices[:val_split_idx]]
    X_val = X[indices[val_split_idx:test_split_idx]]
    X_test = X[indices[test_split_idx:]]
    y_train = y[indices[:val_split_idx]]
    y_val = y[indices[val_split_idx:test_split_idx]]
    y_test = y[indices[test_split_idx:]]

    return X_train, X_val, X_test, y_train, y_val, y_test


def sampling_stratified(
    X: np.ndarray,
    y: np.ndarray,
    regression: bool = False,
    samples_per_class: bool = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Stratified sampling of a dataset.

    This function can be used for both regression and classification problems.

    Parameters
    ----------
    X : np.ndarray
        The input data.

    y : np.ndarray
        The target data.

    regression : bool, optional
        Whether the dataset is for regression or classification.
        Default: False.

    samples_per_class : int, optional
        The number of samples to take per class.
        If None, takes the same number of samples per class as the smallest class.
        Default: None.

    Returns:
    --------
    Tuple[np.ndarray, np.ndarray]
        (X_stratified, y_stratified) : The stratified input and target data.
    """
    assert isinstance(X, np.ndarray), "X should be a numpy array."
    assert isinstance(y, np.ndarray), "y should be a numpy array."
    assert X.shape[0] == y.shape[0], "X and y should be the same axis=0 size."
    assert samples_per_class is None or samples_per_class > 0, "samples_per_class should be greater than 0."

    y_classes = y

    if regression:
        assert y.dtype in [np.float64, np.float32, np.float16], "y should be a float array for regression."

        y_classes = np.digitize(y, np.percentile(y, [10, 20, 30, 40, 50, 60, 70, 80, 90]))

    unique_classes, counts = np.unique(y_classes, return_counts=True)

    if samples_per_class is None:
        samples_per_class = counts.min()

    stratified_indices = np.array([], dtype=int)

    for class_ in unique_classes:
        class_indices = np.where(y_classes == class_)[0]
        selected_indices = np.random.choice(class_indices, samples_per_class, replace=False)
        stratified_indices = np.concatenate((stratified_indices, selected_indices))

    return X[stratified_indices], y[stratified_indices]


def sampling_random(
    X: np.ndarray,
    y: np.ndarray,
    samples: Union[int, float],
) -> Tuple[np.ndarray, np.ndarray]:
    """ 
    Random sampling of a dataset.

    Parameters
    ----------
    X : np.ndarray
        The input data.

    y : np.ndarray
        The target data.

    samples : int or float
        The number of samples to take.
        If int, the number of samples to take.
        If float, the proportion of samples to take.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        (X_random, y_random) : The random input and target data.
    """
    assert isinstance(X, np.ndarray), "X should be a numpy array."
    assert isinstance(y, np.ndarray), "y should be a numpy array."
    assert X.shape[0] == y.shape[0], "X and y should be the same axis=0 size."
    assert samples > 0, "samples should be greater than 0."
    assert isinstance(samples, (int, float)), "samples should be an int or float."

    if isinstance(samples, float) and samples <= 1:
        samples = int(X.shape[0] * samples)

    random_indices = np.random.choice(np.arange(X.shape[0]), samples, replace=False)

    return X[random_indices], y[random_indices]


# TODO: add a function to split a dataset into k folds
# TODO: add a function to split a dataset into k folds stratified by class



"""
### Convert geometry composition. ###

Convert geometries from multiparts and singleparts and vice versa.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import ogr

# Internal
from buteo.utils import (
    utils_base,
    utils_gdal,
    utils_path,
    utils_io,
)
from buteo.vector import core_vector
from buteo.vector.metadata import _vector_to_metadata


def _singlepart_to_multipart(
    vector: Union[str, ogr.DataSource],
    out_path: Optional[str] = None,
    overwrite: bool = True,
    add_index: bool = True,
    process_layer: int = -1,
) -> str:
    """ Internal. """
    assert isinstance(vector, (ogr.DataSource, str)), "Invalid input vector"
    assert utils_gdal._check_is_vector(vector), "Invalid input vector"

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector)

    assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), "Invalid output path"

    ref = core_vector._vector_open(vector)

    out_format = utils_gdal._get_path_from_dataset(out_path, dataset_type="vector")
    driver = ogr.GetDriverByName(out_format)

    utils_path._delete_if_required(out_path, overwrite)

    metadata = _vector_to_metadata(ref)
    destination = driver.CreateDataSource(out_path)

    for index, layer_meta in enumerate(metadata["layers"]):
        if process_layer != -1 and index != process_layer:
            continue

        name = layer_meta["layer_name"]
        geom = layer_meta["column_geom"]

        sql = f"SELECT ST_Collect({geom}) AS geom FROM {name};"

        result = ref.ExecuteSQL(sql, dialect="SQLITE")
        destination.CopyLayer(result, name, ["OVERWRITE=YES"])

    if add_index:
        core_vector.vector_add_index(destination)

    destination.FlushCache()

    return out_path


def _multipart_to_singlepart(
    vector: Union[str, ogr.DataSource],
    out_path: Optional[str] = None,
    copy_attributes: bool = False,
    overwrite: bool = True,
    add_index: bool = True,
    process_layer: int = -1,
    verbose: bool = False,
) -> str:
    """ Internal. """
    assert isinstance(vector, (ogr.DataSource, str)), "Invalid input vector"
    assert utils_gdal._check_is_vector(vector), "Invalid input vector"

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector)

    assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), "Invalid output path"

    ref = core_vector._vector_open(vector)

    out_format = utils_gdal._get_path_from_dataset(out_path, dataset_type="vector")
    driver = ogr.GetDriverByName(out_format)

    utils_path._delete_if_required(out_path, overwrite)

    metadata = _vector_to_metadata(ref)

    utils_path._delete_if_required(out_path, overwrite)

    destination = driver.CreateDataSource(out_path)

    for index, layer_meta in enumerate(metadata["layers"]):
        if process_layer != -1 and index != process_layer:
            continue

        if verbose == 1:
            layer_name = layer_meta["layer_name"]
            print(f"Splitting layer: {layer_name}")

        target_unknown = False

        if layer_meta["geom_type_ogr"] == 4:  # MultiPoint
            target_type = 1  # Point
        elif layer_meta["geom_type_ogr"] == 5:  # MultiLineString
            target_type = 2  # LineString
        elif layer_meta["geom_type_ogr"] == 6:  # MultiPolygon
            target_type = 3  # Polygon
        elif layer_meta["geom_type_ogr"] == 1004:  # MultiPoint (z)
            target_type = 1001  # Point (z)
        elif layer_meta["geom_type_ogr"] == 1005:  # MultiLineString (z)
            target_type = 1002  # LineString (z)
        elif layer_meta["geom_type_ogr"] == 1006:  # MultiPolygon (z)
            target_type = 1003  # Polygon (z)
        elif layer_meta["geom_type_ogr"] == 2004:  # MultiPoint (m)
            target_type = 2001  # Point (m)
        elif layer_meta["geom_type_ogr"] == 2005:  # MultiLineString (m)
            target_type = 2002  # LineString (m)
        elif layer_meta["geom_type_ogr"] == 2006:  # MultiPolygon (m)
            target_type = 2003  # Polygon (m)
        elif layer_meta["geom_type_ogr"] == 3004:  # MultiPoint (zm)
            target_type = 3001  # Point (m)
        elif layer_meta["geom_type_ogr"] == 3005:  # MultiLineString (zm)
            target_type = 3002  # LineString (m)
        elif layer_meta["geom_type_ogr"] == 3006:  # MultiPolygon (zm)
            target_type = 3003  # Polygon (m)
        else:
            target_unknown = True
            target_type = layer_meta["geom_type_ogr"]

        destination_layer = destination.CreateLayer(
            layer_meta["layer_name"], layer_meta["projection_osr"], target_type
        )
        layer_defn = destination_layer.GetLayerDefn()
        field_count = layer_meta["field_count"]

        original_target = ref.GetLayerByIndex(index)
        feature_count = original_target.GetFeatureCount()

        if copy_attributes:
            first_feature = original_target.GetNextFeature()
            original_target.ResetReading()

            if verbose == 1:
                print("Creating attribute fields")

            for field_id in range(field_count):
                field_defn = first_feature.GetFieldDefnRef(field_id)

                fname = field_defn.GetName()
                ftype = field_defn.GetTypeName()
                fwidth = field_defn.GetWidth()
                fprecision = field_defn.GetPrecision()

                if ftype == "String" or ftype == "Date":
                    fielddefn = ogr.FieldDefn(fname, ogr.OFTString)
                    fielddefn.SetWidth(fwidth)
                elif ftype == "Real":
                    fielddefn = ogr.FieldDefn(fname, ogr.OFTReal)
                    fielddefn.SetWidth(fwidth)
                    fielddefn.SetPrecision(fprecision)
                else:
                    fielddefn = ogr.FieldDefn(fname, ogr.OFTInteger)

                destination_layer.CreateField(fielddefn)

        for _ in range(feature_count):
            feature = original_target.GetNextFeature()
            geom = feature.GetGeometryRef()

            if target_unknown:
                out_feat = ogr.Feature(layer_defn)
                out_feat.SetGeometry(geom)

                if copy_attributes:
                    for field_id in range(field_count):
                        values = feature.GetField(field_id)
                        out_feat.SetField(field_id, values)

                destination_layer.CreateFeature(out_feat)

            for geom_part in geom:
                out_feat = ogr.Feature(layer_defn)
                out_feat.SetGeometry(geom_part)

                if copy_attributes:
                    for field_id in range(field_count):
                        values = feature.GetField(field_id)
                        out_feat.SetField(field_id, values)

                destination_layer.CreateFeature(out_feat)

            if verbose:
                utils_base.progress(_, feature_count - 1, "Splitting.")

    if add_index:
        core_vector.vector_add_index(destination)

    return out_path


def vector_singlepart_to_multipart(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    out_path: Optional[str] = None,
    add_index: bool = True,
    process_layer: int = -1,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    overwrite: bool = True,
    allow_lists: bool = True,
) -> Union[str, List[str]]:
    """
    Converts a singlepart vector to multipart.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        The vector(s) to convert.

    out_path : Optional[str], optional
        The path(s) to the output vector. If None a memory output is produced. Default: None

    add_index : bool, optional
        Add an geospatial index to the output vector. Default: True

    process_layer : int, optional
        The layer index to process. Default: -1

    prefix : str, optional
        The prefix to add to the layer name. Default: ""

    suffix : str, optional
        The suffix to add to the layer name. Default: ""
    
    add_uuid : bool, optional
        Add a UUID field to the output vector. Default: False

    overwrite : bool, optional
        Overwrite the output vector if it already exists. Default: True

    allow_lists : bool, optional
        Allow lists of vectors as input. Default: True

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the output vector(s).
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(add_index, [bool], "add_index")
    utils_base._type_check(process_layer, [int], "process_layer")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(allow_lists, [bool], "allow_lists")

    if not allow_lists and isinstance(vector, list):
        raise ValueError("Vector cannot be a list when allow_lists is False.")

    vector_list = utils_base._get_variable_as_list(vector)

    assert utils_gdal._check_is_vector_list(vector_list), f"Vector is not a list of vectors. {vector_list}"

    path_list = utils_io._get_output_paths(
        vector_list,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
    )

    output = []
    for index, in_vector in enumerate(vector_list):
        output.append(
            _singlepart_to_multipart(
                in_vector,
                out_path=path_list[index],
                overwrite=overwrite,
                add_index=add_index,
                process_layer=process_layer,
            )
        )

    if isinstance(vector, list):
        return output[0]

    return output


def vector_multipart_to_singlepart(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    out_path: Optional[str] = None,
    overwrite: bool = True,
    add_index: bool = True,
    process_layer: int = -1,
    verbose: bool = False,
    prefix: str ="",
    suffix: str = "",
    add_uuid: bool = False,
    allow_lists: bool = True,
) -> Union[str, List[str]]:
    """
    Converts a multipart vector to singlepart.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        The vector(s) to convert.

    out_path : Optional[str], optional
        The path(s) to the output vector. If None a memory output is produced. Default: None

    overwrite : bool, optional
        Overwrite the output vector if it already exists. Default: True

    add_index : bool, optional
        Add an geospatial index to the output vector. Default: True

    process_layer : int, optional
        The layer index to process. Default: -1

    verbose : bool, optional
        Print progress. Default: False

    prefix : str, optional
        The prefix to add to the layer name. Default: ""

    suffix : str, optional
        The suffix to add to the layer name. Default: ""

    add_uuid : bool, optional
        Add a UUID field to the output vector. Default: False

    allow_lists : bool, optional
        Allow lists of vectors as input. Default: True

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the output vector(s).
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(add_index, [bool], "add_index")
    utils_base._type_check(process_layer, [int], "process_layer")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(allow_lists, [bool], "allow_lists")

    if not allow_lists and isinstance(vector, list):
        raise ValueError("Vector cannot be a list when allow_lists is False.")

    vector_list = utils_base._get_variable_as_list(vector)

    assert utils_gdal._check_is_vector_list(vector_list), f"Vector is not a list of vectors. {vector_list}"

    path_list = utils_io._get_output_paths(
        vector_list,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
    )

    output = []
    for index, in_vector in enumerate(vector_list):
        output.append(
            _multipart_to_singlepart(
                in_vector,
                out_path=path_list[index],
                overwrite=overwrite,
                add_index=add_index,
                process_layer=process_layer,
                verbose=verbose,
            )
        )

    if isinstance(vector, list):
        return output[0]

    return output



"""
### Encoding spatial values  ###
"""
# Standard library
from typing import Tuple, Union, List
# External
import numpy as np
from numba import jit, prange



@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latitude(lat: float) -> np.ndarray:
    """
    Encode a latitude value into a two-element numpy array of sine and cosine components.

    Parameters
    ----------
    lat : float
        The latitude value to encode. Must be in the range [-90, 90].

    Returns
    -------
    np.ndarray
        A two-element numpy array containing the sine and cosine components of the encoded latitude.
        The sine and cosine values are scaled to the range [0, 1].
    """
    lat_adj = lat + 90.0
    lat_max = 180

    encoded_sin = ((np.sin(2 * np.pi * (lat_adj / lat_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lat_adj / lat_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_longitude(lng: float) -> np.ndarray:
    """
    Encode a longitude value into a two-element numpy array of sine and cosine components.

    Parameters
    ----------
    lng : float
        The longitude value to encode. Must be in the range [-180, 180].

    Returns
    -------
    np.ndarray
        A two-element numpy array containing the sine and cosine components of the encoded longitude.
        The sine and cosine values are scaled to the range [0, 1].
    """
    lng_adj = lng + 180.0
    lng_max = 360

    encoded_sin = ((np.sin(2 * np.pi * (lng_adj / lng_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lng_adj / lng_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latlng(latlng: Union[np.ndarray, Tuple[float, float], List[float]]) -> np.ndarray:
    """
    Encode a latitude-longitude coordinate into a four-element numpy array of sine and cosine components.

    Parameters
    ----------
    latlng : Union[np.ndarray, Tuple[float, float], List[float]]
        A tuple, list, or numpy array containing the latitude and longitude values to encode.
        The latitude value must be in the range [-90, 90] and the longitude value must be in the range [-180, 180].

    Returns
    -------
    np.ndarray
        A four-element numpy array containing the sine and cosine components of the encoded latitude and longitude.
        The sine and cosine values are scaled to the range [0, 1].
    """
    lat = latlng[0]
    lng = latlng[1]

    encoded_lat = encode_latitude(lat)
    encoded_lng = encode_longitude(lng)

    return np.concatenate((encoded_lat, encoded_lng)).astype(np.float32, copy=False)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latlngs(latlngs: np.ndarray) -> np.ndarray:
    """
    Encode multiple latitude-longitude coordinates into a numpy array of sine and cosine components.

    Parameters
    ----------
    latlngs : np.ndarray
        A numpy array containing the latitude-longitude coordinates to encode.
        The shape of the array should be (n, 2), (m, n, 2), or (p, q, n, 2), where n=2 represents the latitude
        and longitude values. The latitude value must be in the range [-90, 90] and the longitude value must be
        in the range [-180, 180].

    Returns
    -------
    np.ndarray
        A numpy array containing the sine and cosine components of the encoded latitude and longitude values.
        The shape of the returned array will depend on the shape of the input array, with the last dimension
        being 4 (two sine and two cosine values).

    Raises
    ------
    ValueError
        If the input array has more than 3 dimensions or the shape of the array is not (n, 2), (m, n, 2),
        or (p, q, n, 2).
    """
    if latlngs.ndim == 1:
        encoded_latlngs = np.apply_along_axis(encode_latlng, 0, latlngs)
    elif latlngs.ndim == 2:
        encoded_latlngs = np.apply_along_axis(encode_latlng, 1, latlngs)
    elif latlngs.ndim == 3:
        rows = latlngs.shape[0]
        cols = latlngs.shape[1]

        output_shape = (rows, cols, 4)
        encoded_latlngs = np.zeros(output_shape, dtype=np.float32)

        for i in prange(rows):
            for j in range(cols):
                latlng = latlngs[i, j]
                encoded_latlngs[i, j] = encode_latlng(latlng)
    else:
        raise ValueError(
            f"The input array must have 1, 2 or 3 dimensions, not {latlngs.ndim}"
        )

    return encoded_latlngs


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latitude(
    encoded_sin: float,
    encoded_cos: float,
) -> float:
    """
    Decode an encoded sine and cosine value to the original latitude value.

    Parameters
    ----------
    encoded_sin : float
        The sine component of the encoded latitude value.
        Must be in the range [0, 1].

    encoded_cos : float
        The cosine component of the encoded latitude value.
        Must be in the range [0, 1].

    Returns
    -------
    float
        The original latitude value. The decoded latitude value is in the range [-90, 90].

    """
    lat_max = 180
    lat_max_half = lat_max / 2.0

    # Calculate the sin and cos values from the encoded values
    sin_val = (2 * encoded_sin) - 1
    cos_val = (2 * encoded_cos) - 1

    # Calculate the latitude adjustment
    lat_adj = np.arctan2(sin_val, cos_val)

    # Convert the adjusted latitude to the original latitude value
    sign = np.sign(lat_adj)
    sign_adj = np.where(sign == 0, 1, sign) * lat_max_half

    lat = ((lat_adj / (2 * np.pi)) * lat_max) - sign_adj
    lat = np.where(lat == -lat_max_half, lat_max_half, lat)

    return lat


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_longitude(
    encoded_sin: float,
    encoded_cos: float,
) -> float:
    """
    Decode an encoded sine and cosine value to the original longitude value.

    Parameters
    ----------
    encoded_sin : float
        The sine component of the encoded longitude value.
        Must be in the range [0, 1].

    encoded_cos : float
        The cosine component of the encoded longitude value.
        Must be in the range [0, 1].

    Returns
    -------
    float
        The original longitude value. The decoded longitude value is in the range [-180, 180].
    """
    lng_max = 360
    lng_max_half = lng_max / 2.0

    # Calculate the sin and cos values from the encoded values
    sin_val = (2 * encoded_sin) - 1
    cos_val = (2 * encoded_cos) - 1

    # Calculate the longitude adjustment
    lng_adj = np.arctan2(sin_val, cos_val)

    # Convert the adjusted longitude to the original longitude value
    sign = np.sign(lng_adj)
    sign_adj = np.where(sign == 0, 1, sign) * lng_max_half

    lng = ((lng_adj / (2 * np.pi)) * lng_max) - sign_adj

    lng = np.where(lng == -lng_max_half, lng_max_half, lng)

    return lng


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latlng(encoded_latlng: Union[np.ndarray, List[Union[float, int]]]) -> np.ndarray:
    """
    Decode an encoded latitude-longitude coordinate to the original latitude-longitude values.

    Parameters
    ----------
    encoded_latlng : Union[np.ndarray, List[Union[float, int]]]
        A numpy array or list containing the encoded latitude-longitude coordinate to decode.
        The shape of the array should be (4,), with the first two values representing the encoded latitude
        and the last two values representing the encoded longitude. The encoded sine and cosine values
        must be in the range [0, 1].

    Returns
    -------
    np.ndarray
        A numpy array containing the original latitude and longitude values. The shape of the returned
        array will be (2,) with the latitude value at index 0 and the longitude value at index 1.
    """
    lat = decode_latitude(encoded_latlng[0], encoded_latlng[1])
    lng = decode_longitude(encoded_latlng[2], encoded_latlng[3])

    return np.array([lat, lng], dtype=np.float32)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latlngs(encoded_latlngs: Union[np.ndarray, List[Union[float, int]]]) -> np.ndarray:
    """
    Decode multiple encoded latitude-longitude coordinates to the original latitude-longitude values.

    Parameters
    ----------
    encoded_latlngs : Union[np.ndarray, List[Union[float, int]]]
        A numpy array or list containing the encoded latitude-longitude coordinates to decode.
        The shape of the array should be (n, 4), where n represents the number of encoded coordinates.
        The first two values of each encoded coordinate represent the encoded latitude and the last two
        values represent the encoded longitude. The encoded sine and cosine values must be in the range [0, 1].

    Returns
    -------
    np.ndarray
        A numpy array containing the original latitude-longitude values. The shape of the returned
        array will be (n, 2), where n represents the number of encoded coordinates. The latitude value
        will be at index 0 and the longitude value will be at index 1.
    """
    latlngs = np.apply_along_axis(decode_latlng, 1, encoded_latlngs)
    return latlngs


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_width(
    lng: float,
    lng_max: float,
) -> float:
    """
    Encode the width of a rectangle to be used as input to the model.

    Parameters
    ----------
    lng : float
        The longitude value representing the width of the rectangle.
        Must be in the range [-180, 180].

    lng_max : float
        The maximum longitude value. This is typically 360 for the full longitude range.

    Returns
    -------
    np.ndarray
        A numpy array containing the encoded width of the rectangle.
        The encoded width consists of two values: the encoded sine and cosine values.
        Both values will be in the range [0, 1].
    """
    encoded_sin = ((np.sin(2 * np.pi * (lng / lng_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lng / lng_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_arr_position(arr: np.ndarray) -> np.ndarray:
    """
    Fast encoding of a 2D numpy array of coordinates where the width is cyclical.
    Very useful for global maps.

    Parameters
    ----------
    arr : np.ndarray
        A 2D numpy array of coordinates to be encoded. The first dimension represents the columns
        and the second dimension represents the rows. The shape of the array should be (m, n), where
        m represents the number of columns and n represents the number of rows.

    Returns
    -------
    np.ndarray
        A numpy array containing the encoded coordinates. The shape of the returned array will be (m, n, 3),
        where m represents the number of columns, n represents the number of rows, and 3 represents the
        number of encoded values for each coordinate. The first two values of each encoded coordinate
        represent the encoded width, while the last value represents the encoded column.
    """
    result = np.zeros((arr.shape[0], arr.shape[1], 3), dtype=np.float32)

    col_end = arr.shape[0] - 1
    row_end = arr.shape[1] - 1

    col_range = np.arange(0, arr.shape[0]).astype(np.float32)
    row_range = np.arange(0, arr.shape[1]).astype(np.float32)

    col_encoded = np.zeros((col_range.shape[0], 1), dtype=np.float32)
    row_encoded = np.zeros((row_range.shape[0], 2), dtype=np.float32)

    for col in prange(col_range.shape[0]):
        col_encoded[col, :] = col_range[col] / col_end

    for row in prange(row_range.shape[0]):
        row_encoded[row, :] = encode_width(row_range[row], row_end)

    for col in prange(arr.shape[0]):
        for row in range(arr.shape[1]):
            result[col, row, 0] = row_encoded[row, 0]
            result[col, row, 1] = row_encoded[row, 1]
            result[col, row, 2] = col_encoded[col, 0]

    return result



"""
### Split functions for vector layers. ###

Dissolve vectors by attributes or geometry.
"""

# Standard library
import sys; sys.path.append("../../")
import os
from typing import Optional, Union
from uuid import uuid4

# External
from osgeo import ogr

# Internal
from buteo.utils import (
    utils_base,
    utils_path,
)
from buteo.vector import core_vector


def vector_split_by_fid(
    vector: Union[str, ogr.DataSource],
    out_folder: Optional[str] = None,
    output_layer_idx: bool = False,
    prefix: str = "",
    suffix: str = "",
) -> str:
    """
    Split a vector by feature id.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource]
        Input vector.

    out_folder : Optional[str], optional
        Output folder, default: None.

    prefix : str, optional
        Prefix for output files, default: "".

    suffix : str, optional
        Suffix for output files, default: "".

    Returns
    -------
    str
        Output paths.
    """
    utils_base._type_check(vector, [ogr.DataSource, str], "vector")
    utils_base._type_check(out_folder, [str, None], "out_folder")
    assert utils_path._check_dir_exists(out_folder) or out_folder is None, "out_folder does not exist."

    opened = core_vector.vector_open(vector, writeable=False)
    metadata = core_vector._get_basic_metadata_vector(opened)

    out_paths = []

    for layer_index in range(metadata["layer_count"]):
        layer = opened.GetLayerByIndex(layer_index)
        layer.ResetReading()

        feature = layer.GetNextFeature()
        while feature:
            feature_id = feature.GetFID()

            if out_folder is None:
                out_path = f"/vsimem/{prefix}{str(uuid4().int)}_{layer_index}_{feature_id}{suffix}.gpkg"
            else:
                if output_layer_idx:
                    out_path = os.path.join(
                        out_folder,
                        f"{prefix}{layer_index}_{feature_id}{suffix}.gpkg",
                    )
                else:
                    out_path = os.path.join(
                        out_folder,
                        f"{prefix}{feature_id}{suffix}.gpkg",
                    )

            out_driver = ogr.GetDriverByName("GPKG")
            out_ds = out_driver.CreateDataSource(out_path)
            out_layer = out_ds.CreateLayer(
                layer.GetName(),
                srs=layer.GetSpatialRef(),
                geom_type=layer.GetGeomType(),
            )

            out_layer.CreateFeature(feature)
            out_layer.SyncToDisk()

            out_ds.FlushCache()
            out_ds, out_layer, feature = (None, None, None)

            out_paths.append(out_path)

            feature = layer.GetNextFeature()

    return out_paths


def vector_split(
    vector: Union[str, ogr.DataSource],
    out_folder: Optional[str] = None,
    output_layer_idx: bool = False,
    prefix: str = "",
    suffix: str = "",
) -> str:
    """
    Split a vector by feature id.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource]
        Input vector.

    out_folder : Optional[str], optional
        Output folder, default: None.

    prefix : str, optional
        Prefix for output files, default: "".

    suffix : str, optional
        Suffix for output files, default: "".

    Returns
    -------
    str
        Output paths.
    """
    utils_base._type_check(vector, [ogr.DataSource, str], "vector")
    utils_base._type_check(out_folder, [str, None], "out_folder")
    assert utils_path._check_dir_exists(out_folder) or out_folder is None, "out_folder does not exist."

    opened = core_vector.vector_open(vector, writeable=False)
    metadata = core_vector._get_basic_metadata_vector(opened)

    out_paths = []

    for layer_index in range(metadata["layer_count"]):
        layer = opened.GetLayerByIndex(layer_index)
        layer.ResetReading()

        feature = layer.GetNextFeature()
        feature_id = 0
        while feature:

            if out_folder is None:
                out_path = f"/vsimem/{prefix}{str(uuid4().int)}_{layer_index}_{feature_id}{suffix}.gpkg"
            else:
                if output_layer_idx:
                    out_path = os.path.join(
                        out_folder,
                        f"{prefix}{layer_index}_{feature_id}{suffix}.gpkg",
                    )
                else:
                    out_path = os.path.join(
                        out_folder,
                        f"{prefix}{feature_id}{suffix}.gpkg",
                    )

            out_driver = ogr.GetDriverByName("GPKG")
            out_ds = out_driver.CreateDataSource(out_path)
            out_layer = out_ds.CreateLayer(
                layer.GetName(),
                srs=layer.GetSpatialRef(),
                geom_type=layer.GetGeomType(),
            )

            out_layer.CreateFeature(feature)
            out_layer.SyncToDisk()

            out_ds.FlushCache()
            out_ds, out_layer, feature = (None, None, None)

            out_paths.append(out_path)

            feature = layer.GetNextFeature()
            feature_id += 1

    return out_paths



"""
### Calculate intersections ###

Calculate and tests the intersections between geometries.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, List, Optional

# External
from osgeo import ogr, gdal

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_path,
)
from buteo.vector import core_vector
from buteo.vector.metadata import _vector_to_metadata
from buteo.vector.reproject import _vector_reproject
from buteo.vector.merge import vector_merge_layers


def _vector_intersect(
    vector: Union[str, ogr.DataSource],
    clip_geom: Union[str, ogr.Geometry],
    out_path: Optional[str] = None,
    process_layer: int = 0,
    process_layer_clip: int = 0,
    add_index: bool = True,
    overwrite: bool = True,
    return_bool: bool = False,
) -> Union[str, bool]:
    """ Internal. """
    assert isinstance(vector, ogr.DataSource), f"Invalid input vector: {vector}"
    assert utils_gdal._check_is_vector(vector), f"Invalid input vector: {vector}"

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector, suffix="_intersect")

    assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), "Invalid output path."

    match_projection = _vector_reproject(clip_geom, vector)
    geometry_to_clip = core_vector._vector_open(match_projection)

    merged = core_vector._vector_open(vector_merge_layers([vector, match_projection]))

    if add_index:
        core_vector.vector_add_index(merged)

    vector_metadata = _vector_to_metadata(vector)
    vector_layername = vector_metadata["layers"][process_layer]["layer_name"]
    vector_geom_col = vector_metadata["layers"][process_layer]["column_geom"]

    clip_geom_metadata = _vector_to_metadata(geometry_to_clip)
    clip_geom_layername = clip_geom_metadata["layers"][process_layer_clip]["layer_name"]
    clip_geom_col = clip_geom_metadata["layers"][process_layer_clip]["column_geom"]

    if return_bool:
        sql = f"SELECT A.* FROM '{vector_layername}' A, '{clip_geom_layername}' B WHERE ST_INTERSECTS(A.{vector_geom_col}, B.{clip_geom_col});"
    else:
        sql = f"SELECT A.* FROM '{vector_layername}' A, '{clip_geom_layername}' B WHERE ST_INTERSECTS(A.{vector_geom_col}, B.{clip_geom_col});"

    result = merged.ExecuteSQL(sql, dialect="SQLITE")

    if return_bool:
        if result.GetFeatureCount() == 0:
            return False
        else:
            return True

    driver = ogr.GetDriverByName(utils_gdal._get_vector_driver_name_from_path(out_path))
    destination = driver.CreateDataSource(out_path)
    destination.CopyLayer(result, vector_layername, ["OVERWRITE=YES"])

    if destination is None:
        raise RuntimeError("Error while running intersect.")

    destination.FlushCache()

    return out_path


def vector_intersect(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    clip_geom: Union[str, ogr.Geometry, List[Union[str, ogr.Geometry]]],
    out_path: Optional[str] = None,
    process_layer: int = 0,
    process_layer_clip: int = 0,
    add_index: bool = True,
    overwrite: bool = True,
    prefix: str ="",
    suffix: str ="",
    add_uuid: bool = False,
    allow_lists: bool = True,
) -> Union[str, List[str]]:
    """
    Clips a vector to a geometry.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        The vector(s) to intersect.

    clip_geom : Union[str, ogr.Geometry, List[Union[str, ogr.Geometry]]]
        The geometry to intersect the vector(s) with.

    out_path : Optional[str], optional
        The path(s) to save the clipped vector(s) to. Default: None

    process_layer : int, optional
        The layer to process in the vector(s). Default: 0

    process_layer_clip : int, optional
        The layer to process in the clip geometry. Default: 0

    add_index : bool, optional
        Add a geospatial index to the vector(s). Default: True

    overwrite : bool, optional
        Overwrite the output vector(s) if they already exist. Default: True

    prefix : str, optional
        A prefix to add to the output vector(s). Default: ""

    suffix : str, optional
        A suffix to add to the output vector(s). Default: ""

    add_uuid : bool, optional
        Add a uuid to the output vector(s). Default: False

    allow_lists : bool, optional
        Allow lists as input. Default: True

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the intersected vector(s).
    """
    utils_base._type_check(vector, [ogr.DataSource, str, list], "vector")
    utils_base._type_check(clip_geom, [ogr.DataSource, gdal.Dataset, str, list, tuple], "clip_geom")
    utils_base._type_check(out_path, [str], "out_path", allow_none=True)
    utils_base._type_check(process_layer, [int], "process_layer")
    utils_base._type_check(process_layer_clip, [int], "process_layer_clip")
    utils_base._type_check(add_index, [bool], "add_index")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(allow_lists, [bool], "allow_lists")

    if not allow_lists and isinstance(vector, list):
        raise ValueError("Lists are not allowed when allow_lists is False.")

    vector_list = utils_base._get_variable_as_list(vector)

    assert utils_gdal._check_is_vector_list(vector_list), f"Invalid input vector: {vector_list}"

    path_list = utils_io._get_output_paths(
        vector_list,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )

    assert utils_path._check_is_valid_output_filepath(path_list, overwrite=overwrite), "Invalid output path generated."

    output = []
    for index, in_vector in enumerate(vector_list):
        output.append(
            _vector_intersect(
                in_vector,
                clip_geom,
                out_path=path_list[index],
                process_layer=process_layer,
                process_layer_clip=process_layer_clip,
                add_index=add_index,
                overwrite=overwrite,
                return_bool=False,
            )
        )

    if isinstance(vector, list):
        return output

    return output[0]




"""
### Generic utility functions ###

Functions that make interacting with the toolbox easier.
"""

# Standard Library
import os
import gc
import sys
import shutil
from datetime import datetime
from typing import List



def _print_progress(
    count: int,
    total: int,
    name: str = "Processing",
) -> None:
    """
    Print a progress bar.

    `progress(10, 100, "Processing..")`

    Parameters
    ----------
    count : int
        The current count.

    total : int
        The total count.

    name : str, optional.
        The name of the process. Default: "Processing".

    Returns
    -------
    None
    """
    assert isinstance(count, int), "count must be an integer."
    assert isinstance(total, int), "total must be an integer."
    assert isinstance(name, str), "name must be a string."

    sys.stdout.flush()

    try:
        bar_len = os.get_terminal_size().columns - 24
    except RuntimeError:
        bar_len = shutil.get_terminal_size().columns - 24

    filled_len = int(round(bar_len * count / float(total)))
    display_name = name[:10] + "..: "

    progress_bar = "█" * filled_len + "." * (bar_len - filled_len)

    percents = round(100.0 * count / float(total), 1)

    if percents >= 100.0:
        percents = 100.0

    if count == total:
        sys.stdout.write(f"{display_name}[{progress_bar}] {percents} %\r")
        sys.stdout.flush()
        print("")
        return None
    else:
        sys.stdout.write(f"{display_name}[{progress_bar}] {percents} %\r")
        sys.stdout.flush()

    return None


def _get_timing(
    before: datetime,
    print_msg: bool = True,
) -> str:
    """
    Get the time elapsed since the given time.

    Parameters
    ----------
    before : datetime
        The time to compare to.
    
    print_msg : bool, optional.
        If True, print the message. Default: True.

    Returns
    -------
    str
        The message.

    Examples
    --------
    ```python
    >>> from datetime import datetime
    >>> from buteo.utils import timing

    >>> before = datetime.now()
    >>> # Do something
    >>> timing(before)
    "Processing took: 0h 2m 3.1s"
    ```
    """
    assert isinstance(before, datetime), "before must be a datetime object."

    after = datetime.now()
    dif = (after - before).total_seconds()

    hours = int(dif / 3600)
    minutes = int((dif % 3600) / 60)
    seconds = f"{dif % 60:.2f}"

    message = f"Processing took: {hours}h {minutes}m {seconds}s"

    if print_msg:
        print(message)

    return message


def _get_folder_size(
    start_path: str = ".",
    rough: bool = True,
) -> int:
    """
    Get the size of a folder.

    Parameters
    ----------
    start_path : str, optional.
        The path to the folder. Default: ".".
    
    rough : bool, optional.
        If True, return the size in MB. Default: True.

    Returns
    -------
    int
        The size of the folder in bytes or MB.
    """
    assert isinstance(start_path, str), "start_path must be a string."
    assert isinstance(rough, bool), "rough must be a boolean."

    total_size = 0
    for dirpath, _dirnames, filenames in os.walk(start_path):
        for file in filenames:
            file_path = os.path.join(dirpath, file)

            if not os.path.islink(file_path): # skip if symbolic link
                total_size += os.path.getsize(file_path)

    if rough is True:
        return total_size >> 20

    return total_size


def _force_garbage_collect_all(
    delete_functions: bool = True,
) -> None:
    """
    Clears the memory by deleting all objects in the main namespace.
    Very aggresive. Use with caution.
    
    Parameters
    ----------
    delete_functions : bool, optional.
        If True, delete functions as well. Default: True.

    Returns
    -------
    None
    """
    # Get a list of all objects
    all_objects = sys.modules['__main__'].__dict__.copy()

    # Iterate over the objects and delete them if possible
    for key, value in all_objects.items():
        if key in ["sys", "os", "gc", "buteo"]:
            continue
        if key.startswith("__"):
            continue  # Skip built-in objects
        if not delete_functions and callable(value):
            continue

        del sys.modules['__main__'].__dict__[key]  # Remove the object from the namespace

    # Collect garbage
    gc.collect()


def split_number(num: int, parts: int) -> List[int]:
    """
    Splits a number into parts of equal size.
    The remainder is distributed starting on the left.

    ```python
    >>> split_number(10, 3)
    [4, 3, 3]
    >>> split_number(10, 4)
    [3, 3, 2, 2]
    ```

    Parameters
    ----------
    num : int
        The number to split.

    parts : int
        The number of parts to split the number into.

    Returns
    -------
    List[int]
        A list of the parts.
    """
    assert isinstance(num, int), "num must be an integer."
    assert isinstance(parts, int), "parts must be an integer."
    assert num >= parts, "num must be larger than parts."

    # Calculate the base and remainder
    base_num = num // parts
    remainder = num % parts
    
    # Create the result list
    result = [base_num] * parts
    
    # Add the remainder to the elements from the left
    for i in range(remainder):
        result[i] += 1
    
    return result



"""
### Generic utility functions ###

Functions that make interacting with the toolbox easier.
"""
from typing import Tuple

# External
import numpy as np
from numba import jit, prange


def channel_first_to_last(arr: np.ndarray) -> np.ndarray:
    """
    Converts a numpy array from channel first to channel last format.
    `(-batch-, channel, height, width)` -> `(-batch-, height, width, channel)`
    
    If 4D, it is assumed that the input array is in batch, channel, height, width format.
    If 3D, it is assumed that the input array is in channel, height, width format.

    Parameters
    ----------
    arr : np.ndarray
        The array to convert.

    Returns
    -------
    np.ndarray
        The converted array.
    """
    if arr.ndim not in [3, 4]:
        raise ValueError("Input array should be 3 or 4-dimensional with shape (-batch-, channels, height, width)")

    # Swap the axes to change from channel first to channel last format
    if arr.ndim == 3:
        arr = np.transpose(arr, (1, 2, 0))
    else:
        arr = np.transpose(arr, (0, 2, 3, 1))

    return arr


def channel_last_to_first(arr: np.ndarray) -> np.ndarray:
    """
    Converts a numpy array from channel last to channel first format.

    `(-batch-, height, width, channel)` -> `(-batch-, channel, height, width)`
    
    If 4D, it is assumed that the input array is in batch, channel, height, width format.
    If 3D, it is assumed that the input array is in channel, height, width format.

    Parameters
    ----------
    arr : np.ndarray
        The array to convert.

    Returns
    -------
    np.ndarray
        The converted array.
    """
    if arr.ndim not in [3, 4]:
        raise ValueError("Input array should be 3 or 4-dimensional with shape (-batch-, height, width, channels)")

    # Swap the axes to change from channel last to channel first format
    if arr.ndim == 3:
        arr = np.transpose(arr, (2, 0, 1))
    else:
        arr = np.transpose(arr, (0, 3, 1, 2))

    return arr


@jit(nopython=True)
def _create_grid(range_rows, range_cols):
    """
    Create a grid of rows and columns.
    
    Parameters
    ----------
    range_rows : np.ndarray
        The rows to create the grid from.

    range_cols : np.ndarray
        The columns to create the grid from.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The rows and columns grid.
    """
    rows_grid = np.zeros((len(range_rows), len(range_cols)), dtype=np.int64)
    cols_grid = np.zeros((len(range_rows), len(range_cols)), dtype=np.int64)

    for i in range(len(range_rows)):
        for j in range(len(range_cols)):
            cols_grid[i, j] = range_rows[j]
            rows_grid[i, j] = range_cols[i]

    return rows_grid, cols_grid


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latitude(lat):
    """ Latitude goes from -90 to 90 """
    lat_adj = lat + 90.0
    lat_max = 180

    encoded_sin = ((np.sin(2 * np.pi * (lat_adj / lat_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lat_adj / lat_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_longitude(lng):
    """ Longitude goes from -180 to 180 """
    lng_adj = lng + 180.0
    lng_max = 360

    encoded_sin = ((np.sin(2 * np.pi * (lng_adj / lng_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lng_adj / lng_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latlng(latlng):
    """
    Encode latitude and longitude values to be used as input to the model.
    """
    lat = latlng[0]
    lng = latlng[1]

    encoded_lat = encode_latitude(lat)
    encoded_lng = encode_longitude(lng)

    return np.concatenate((encoded_lat, encoded_lng)).astype(np.float32)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_latlngs(latlngs):
    """ Encode multiple latitude and longitude values. """
    if latlngs.ndim == 1:
        encoded_latlngs = np.apply_along_axis(encode_latlng, 0, latlngs)
    elif latlngs.ndim == 2:
        encoded_latlngs = np.apply_along_axis(encode_latlng, 1, latlngs)
    elif latlngs.ndim == 3:
        rows = latlngs.shape[0]
        cols = latlngs.shape[1]

        output_shape = (rows, cols, 4)
        encoded_latlngs = np.zeros(output_shape, dtype=np.float32)

        for i in prange(rows):
            for j in range(cols):
                latlng = latlngs[i, j]
                encoded_latlngs[i, j] = encode_latlng(latlng)
    else:
        raise ValueError(
            f"The input array must have 1, 2 or 3 dimensions, not {latlngs.ndim}"
        )

    return encoded_latlngs


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latitude(encoded_sin, encoded_cos):
    """
    Decode encoded latitude values to the original latitude value.
    """
    lat_max = 180
    lat_max_half = lat_max / 2.0

    # Calculate the sin and cos values from the encoded values
    sin_val = (2 * encoded_sin) - 1
    cos_val = (2 * encoded_cos) - 1

    # Calculate the latitude adjustment
    lat_adj = np.arctan2(sin_val, cos_val)

    # Convert the adjusted latitude to the original latitude value
    sign = np.sign(lat_adj)
    sign_adj = np.where(sign == 0, 1, sign) * lat_max_half

    lat = ((lat_adj / (2 * np.pi)) * lat_max) - sign_adj
    lat = np.where(lat == -lat_max_half, lat_max_half, lat)

    return lat

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_longitude(encoded_sin, encoded_cos):
    """
    Decode encoded longitude values to the original longitude value.
    """
    lng_max = 360
    lng_max_half = lng_max / 2.0

    # Calculate the sin and cos values from the encoded values
    sin_val = (2 * encoded_sin) - 1
    cos_val = (2 * encoded_cos) - 1

    # Calculate the longitude adjustment
    lng_adj = np.arctan2(sin_val, cos_val)

    # Convert the adjusted longitude to the original longitude value
    sign = np.sign(lng_adj)
    sign_adj = np.where(sign == 0, 1, sign) * lng_max_half

    lng = ((lng_adj / (2 * np.pi)) * lng_max) - sign_adj

    lng = np.where(lng == -lng_max_half, lng_max_half, lng)

    return lng


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latlng(encoded_latlng):
    """
    Decode encoded latitude and longitude values to the original values.
    """
    lat = decode_latitude(encoded_latlng[0], encoded_latlng[1])
    lng = decode_longitude(encoded_latlng[2], encoded_latlng[3])

    return np.array([lat, lng], dtype=np.float32)

@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def decode_latlngs(encoded_latlngs):
    """ Decode multiple latitude and longitude values. """
    latlngs = np.apply_along_axis(decode_latlng, 1, encoded_latlngs)
    return latlngs


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, inline="always")
def encode_width(lng, lng_max):
    """ Longitude goes from -180 to 180 """

    encoded_sin = ((np.sin(2 * np.pi * (lng / lng_max)) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * (lng / lng_max)) + 1)) / 2.0

    return np.array([encoded_sin, encoded_cos], dtype=np.float32)


@jit(nopython=True, fastmath=True, cache=True, nogil=True, inline='always')
def single_hue_to_rgb(
    p: float,
    q: float,
    t: float,
) -> float:
    """
    Helper function to convert hue to RGB.

    Args:
        p (float): Intermediate value used for hue to RGB conversion.
        q (float): Intermediate value used for hue to RGB conversion.
        t (float): Hue value.

    Returns:
        float: RGB value.
    """
    if t < 0:
        t += 1
    if t > 1:
        t -= 1
    if t < 1/6:
        return p + (q - p) * 6 * t
    if t < 1/2:
        return q
    if t < 2/3:
        return p + (q - p) * (2/3 - t) * 6
    return p


@jit(nopython=True, fastmath=True, cache=True, nogil=True, inline='always')
def single_hsl_to_rgb(
    h: float,
    s: float,
    l: float,
) -> Tuple[float, float, float]:
    """
    Convert a single HSL color to RGB.

    Args:
        h (float): Hue component.
        s (float): Saturation component.
        l (float): Lightness component.

    Returns:
        Tuple[float, float, float]: Tuple of RGB values (r, g, b).
    """
    if s == 0:
        return l, l, l

    q = l * (1 + s) if l < 0.5 else l + s - l * s
    p = 2 * l - q

    r = single_hue_to_rgb(p, q, h + 1/3)
    g = single_hue_to_rgb(p, q, h)
    b = single_hue_to_rgb(p, q, h - 1/3)

    return r, g, b


@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def hsl_to_rgb(hsl_array: np.ndarray) -> np.ndarray:
    """
    Convert an HSL array to an RGB array.

    Args:
        hsl_array (np.ndarray): Input HSL array with shape (height, width, 3).

    Returns:
        np.ndarray: Output RGB array with shape (height, width, 3).
    """
    assert hsl_array.ndim == 3, "Input array must have 3 dimensions"
    assert hsl_array.shape[-1] == 3, "Input array must have 3 channels"
    assert hsl_array.min() >= 0 and hsl_array.max() <= 1, "Input array must be normalized"

    shape = hsl_array.shape

    rgb_array = np.empty(shape, dtype=np.float32)
    for i in prange(shape[0]):
        for j in range(shape[1]):
            h, s, l = hsl_array[i, j]

            r, g, b = single_hsl_to_rgb(h, s, l)

            if hsl_array.ndim == 3:
                rgb_array[i, j, 0] = r
                rgb_array[i, j, 1] = g
                rgb_array[i, j, 2] = b
            else:
                rgb_array[i, j] = [r, g, b]

    return rgb_array


@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def rgb_to_hsl(rgb_array: np.ndarray) -> np.ndarray:
    """ Convert an RGB array to an HSL array.
    
    Args:
        rgb_array (np.ndarray): Input RGB array with shape (height, width, 3).
    
    Returns:
        np.ndarray: Output HSL array with shape (height, width, 3).
    """
    assert rgb_array.ndim == 3, "Input array must have 3 dimensions"
    assert rgb_array.shape[-1] == 3, "Input array must have 3 channels"
    assert rgb_array.min() >= 0 and rgb_array.max() <= 1, "Input array must be normalized"

    # Get the shape of the input array
    shape = rgb_array.shape

    # Initialize the minimum and maximum arrays
    cmin = np.zeros((shape[0], shape[1]))
    cmax = np.zeros((shape[0], shape[1]))

    # Calculate the minimum and maximum of the RGB values for each pixel
    for i in prange(shape[0]):
        for j in prange(shape[1]):
            cmin[i, j] = np.min(rgb_array[i, j, :])
            cmax[i, j] = np.max(rgb_array[i, j, :])

    # Calculate the difference of the RGB values
    delta = cmax - cmin

    # Initialize the HSL arrays
    hue = np.zeros((shape[0], shape[1]))
    saturation = np.zeros((shape[0], shape[1]))
    luminosity = (cmax + cmin) / 2

    # Initialize the HSL array
    hsl_array = np.zeros((shape[0], shape[1], 3))

    red, green, blue = rgb_array[..., 0], rgb_array[..., 1], rgb_array[..., 2]

    for i in prange(shape[0]):
        for j in prange(shape[1]):
            if delta[i, j] != 0:
                saturation[i, j] = delta[i, j] / (1 - np.abs(2 * luminosity[i, j] - 1))

                if cmax[i, j] == red[i, j]:
                    hue[i, j] = (green[i, j] - blue[i, j]) / delta[i, j] % 6
                elif cmax[i, j] == green[i, j]:
                    hue[i, j] = (blue[i, j] - red[i, j]) / delta[i, j] + 2
                elif cmax[i, j] == blue[i, j]:
                    hue[i, j] = (red[i, j] - green[i, j]) / delta[i, j] + 4

                hue[i, j] = (hue[i, j] * 60) % 360
                if hue[i, j] < 0:
                    hue[i, j] += 360

    # Normalize the hue value to [0, 1]
    hue /= 360

    # Assign the h, s, and l values to the HSL array
    hsl_array[..., 0] = hue
    hsl_array[..., 1] = saturation
    hsl_array[..., 2] = luminosity

    hsl_array = np.clip(hsl_array, 0.0, 1.0)

    return hsl_array
